<?php

namespace OCA\Dematpayslip\BackgroundJob;

use OC\BackgroundJob\TimedJob;
use OCA\Dematpayslip\Db\DematpayslipProcess;
use OCP\AppFramework\Utility\ITimeFactory;
use OCP\Files\IRootFolder;
use OCP\Activity\IManager;
use OCP\IConfig;
use OCP\IDBConnection;
use OCP\IL10N;
use OCP\Notification\IManager as INotificationManager;
use OCP\Share\IManager as IShareManager;
use OCP\IUserManager;

use DateTime;
use DateTimeZone;
use Exception;

use OCA\Dematpayslip\Db\DematpayslipArchiveMapper;
use OCA\Dematpayslip\Db\DematpayslipProcessMapper;
use OCA\Dematpayslip\Db\DematpayslipProcessSkipMapper;
use OCA\Dematpayslip\Db\DematpayslipUserMapper;
use OCA\Dematpayslip\Activity\Provider;
use OCA\Dematpayslip\Notification\Notifier;


    /**
 * Clean up all file locks that are expired for the DB file locking provider
 */
class IndexationProcess extends TimedJob {
    /**
     * Const log levels
     */
    const LOG_SUCCESS = 0;
    const LOG_DEBUG = 1;
    const LOG_WARNING = 2;
    const LOG_ERROR = 3;

    /**
     * Const log modes
     */
    const LOG_MODE_NONE = -1;
    const LOG_MODE_FILE = 0;
    const LOG_MODE_SCREEN = 1;

    /**
     * @var string app name
     */
    private $_appName;

    /**
     * @var IConfig config
     */
    private $_config;

    /**
     * @var IDBConnection database connection
     */
    private $_dbConnection;

    /**
     * @var IL10N lang
     */
    private $_l10n;

    /**
     * @var IManager activity manager
     */
    private $_activityManager;

    /**
     * @var INotificationManager notification manager
     */
    private $_notificationManager;

    /**
     * @var IRootFolder root folder
     */
    private $_rootFolder;

    /**
     * @var IShareManager share manager
     */
    private $_shareManager;

    /**
     * @var ITimeFactory time factory
     */
    private $_timeFactory;

    /**
     * @var DateTimeZone time zone
     */
    private $_timeZone;

    /**
     * @var IUserManager user manager
     */
    private $_userManager;


    /**
     * @var DematpayslipArchiveMapper archive mapper
     */
    private $_dematpayslipArchiveMapper;

    /**
     * @var DematpayslipProcessMapper process mapper
     */
    private $_dematpayslipProcessMapper;

    /**
     * @var DematpayslipProcessSkipMapper process skip mapper
     */
    private $_dematpayslipProcessSkipMapper;

    /**
     * @var DematpayslipUserMapper user mapper
     */
    private $_dematpayslipUserMapper;

    /**
     * @var int log level
     */
    private $_logLevel;

    /**
     * @var int log mode
     */
    private $_logMode;

    /**
     * @var resource log file
     */
    private $_logFile;

    /**
     * Log
     *
     * @param   string  $msg            Log message
     * @param   int     $level          [=1] Log level (Debug by default)
     *
     * @throws  Exception
     */
    private function _log($msg, $level = self::LOG_DEBUG) {
        if ($this->_logMode >= self::LOG_MODE_FILE) {
            if ($this->_logLevel <= $level || $level === self::LOG_SUCCESS) {
                $dateTimeNow = new DateTime('now', $this->_timeZone);
                $logLevel = '';
                switch ($level) {
                    case self::LOG_SUCCESS :
                        $logLevel = 'SUCCESS';
                        break;

                    case self::LOG_DEBUG :
                        $logLevel = 'DEBUG';
                        break;

                    case self::LOG_WARNING :
                        $logLevel = 'WARNING';
                        break;

                    case self::LOG_ERROR :
                        $logLevel = 'ERROR';
                        break;

                    default :
                        break;
                }

                $logMsg = $dateTimeNow->format('Y-m-d_H:i:s') . ' [' . $logLevel . '] : ' . $msg . "\n";

                if ($this->_logMode == self::LOG_MODE_FILE && $this->_logFile !== null) {
                    fwrite($this->_logFile, $logMsg);
                } elseif ($this->_logMode == self::LOG_MODE_SCREEN) {
                    print $logMsg;
                }
            }
        }
    }

    /**
     * IndexationProcess constructor.
     *
     * @param   IConfig                         $config                         Config
     * @param   IDBConnection                   $dbConnection                   DB connection
     * @param   IL10N                           $l10n                           Lang
     * @param   IManager                        $activityManager                Activity manager
     * @param   IRootFolder                     $rootFolder                     Root folder
     * @param   INotificationManager            $notificationManager            Notification manager
     * @param   IShareManager                   $shareManager                   Share manager
     * @param   ITimeFactory                    $timeFactory                    Time factory
     * @param   IUserManager                    $userManager                    User manager
     * @param   DematpayslipArchiveMapper       $dematpayslipArchiveMapper      Archive mapper
     * @param   DematpayslipProcessMapper       $dematpayslipProcessMapper      Process mapper
     * @param   DematpayslipProcessSkipMapper   $dematpayslipProcessSkipMapper  Process skip mapper
     * @param   DematpayslipUserMapper          $dematpayslipUserMapper         User mapper
     */
    public function __construct(IConfig $config, IDBConnection $dbConnection, IL10N $l10n, IManager $activityManager, INotificationManager $notificationManager, IRootFolder $rootFolder, IShareManager $shareManager, ITimeFactory $timeFactory, IUserManager $userManager, DematpayslipArchiveMapper $dematpayslipArchiveMapper, DematpayslipProcessMapper $dematpayslipProcessMapper, DematpayslipProcessSkipMapper $dematpayslipProcessSkipMapper, DematpayslipUserMapper $dematpayslipUserMapper) {
        //print __METHOD__ . "\n";

        $this->_appName = 'dematpayslip';
        $this->_activityManager = $activityManager;
        $this->_config = $config;
        $this->_dbConnection = $dbConnection;
        $this->_l10n = $l10n;
        $this->_notificationManager = $notificationManager;
        $this->_rootFolder = $rootFolder;
        $this->_shareManager = $shareManager;
        $this->_timeFactory = $timeFactory;
        $this->_userManager = $userManager;

        $this->_dematpayslipArchiveMapper = $dematpayslipArchiveMapper;
        $this->_dematpayslipProcessMapper = $dematpayslipProcessMapper;
        $this->_dematpayslipProcessSkipMapper = $dematpayslipProcessSkipMapper;
        $this->_dematpayslipUserMapper = $dematpayslipUserMapper;

        //$timeZoneDefaultName = 'UTC';
        $timeZoneDefaultName = 'Europe/Paris';
        $timeZoneName = $this->_config->getAppValue($this->_appName, 'indexation_cron_time_zone', $timeZoneDefaultName);
        $this->_timeZone = new DateTimeZone($timeZoneName);

        // run once by interval
        $taskDefaultInterval = 3600; // 1h
        $taskInterval = intval($this->_config->getAppValue($this->_appName,'indexation_cron_interval', $taskDefaultInterval));
        //print __METHOD__ . ' : time_zone_name=' . $timeZoneName . ', cron_interval=' . $taskInterval . "\n";
        $this->setInterval($taskInterval);
	}

    /**
     * Run
     *
     * @param   mixed   $arguments  Arguments
     *
     * @throws  Exception
     */
	protected function run($arguments) {
        try {
            $dateTimeNow = new DateTime('now', $this->_timeZone);
            print __METHOD__ . ' : cron run start_time=' . $dateTimeNow->getTimestamp() . ", date=" . $dateTimeNow->format('d/m/Y H:i:s') . "\n";

            // log file
            $logLevel    = $this->_config->getAppValue($this->_appName, 'indexation_cron_log_level', self::LOG_ERROR);
            $logMode     = $this->_config->getAppValue($this->_appName, 'indexation_cron_log_mode', self::LOG_MODE_NONE);
            $logFileName = $this->_config->getAppValue($this->_appName, 'indexation_cron_log_file_name', $this->_appName . '_indexation_cron.log');
            $logFilePath = $this->_config->getSystemValue('logfile');
            if (empty($logFilePath)) {
                $logDirPath = $this->_config->getSystemValue('datadirectory');
            } else {
                $logDirPath = dirname($this->_config->getSystemValue('logfile'));
            }
            $logFilePath = $logDirPath . DIRECTORY_SEPARATOR . $logFileName;
            //print __METHOD__ . ' : level=' . $logLevel . ', mode=' . $logMode . ', file_name=' . $logFileName . ', file_path=' . $logFilePath . "\n";

            $logFile = fopen($logFilePath, 'a+');
            if ($logFile) {
                //$dateTimeNow = new DateTime('now', $this->_timeZone);
                //print __METHOD__ . ' : job start_time=' . $dateTimeNow->getTimestamp() . "\n";

                $this->_logFile = $logFile;
                $this->_logLevel = $logLevel;
                $this->_logMode = $logMode;

                $userId = $this->_config->getAppValue($this->_appName, 'user_id', 'demat');
                // process message list
                $processMsgList = array(
                    'error'   => array(),
                    'success' => array(),
                );
                $this->_log('Indexation process : Begin');

                // find process in indexation step
                $processList = $this->_dematpayslipProcessMapper->findAllLast();
                if (!empty($processList)) {
                    $process = $processList[0];
                    if ($process->getPhase() == DematpayslipProcess::PHASE_ID_INDEXATION) {
                        $process->load($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $this->_shareManager, $this->_userManager, $userId, $this->_dematpayslipProcessMapper, $this->_dematpayslipProcessSkipMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                        $result = $process->indexation($processMsgList);

                        if ($result < 0) {
                            $msgErrorList = $process->getMsgList('error');
                            foreach ($msgErrorList as $msgError) {
                                $this->_log('Error : ' . $msgError, self::LOG_ERROR);
                            }
                        } else {
                            $nbError = count($processMsgList['error']);
                            $nbSuccess = count($processMsgList['success']);
                            if ($nbError > 0 || $nbSuccess > 0) {
                                $nbTotal = $nbError + $nbSuccess;
                                $dateTimeNow = new DateTime('now', $this->_timeZone);

                                // activity
                                $this->_log('Activity add : indexation processed');
                                $event = $this->_activityManager->generateEvent();
                                $event->setApp($this->_appName)
                                    ->setType('job_processed')
                                    ->setAffectedUser($userId)
                                    ->setTimestamp($dateTimeNow->getTimestamp())
                                    ->setSubject(Provider::SUBJECT_ARCHIVES_PROCESSED_ACTIVITY, [$nbSuccess, $nbTotal]);
                                //->setLink($folderLink)
                                //->setObject('folder', $folder->getId(), $folder->getUuid());
                                $this->_activityManager->publish($event);

                                // notification
                                $this->_log('Notification add : indexation processed');
                                $notification = $this->_notificationManager->createNotification();
                                $notification->setApp($this->_appName)
                                    ->setSubject(Notifier::SUBJECT_JOB_PROCESSED_NOTIFY, ['nb_success' => $nbSuccess, 'nb_total' => $nbTotal])
                                    ->setObject('dematpayslip_process', $dateTimeNow->getTimestamp())
                                    ->setUser($userId);
                                //->setIcon($this->_urlGenerator->getAbsoluteURL($this->_urlGenerator->imagePath($this->_appName, 'folder.png')))
                                //->setLink($folderLink);
                                if ($this->_notificationManager->getCount($notification) === 0) {
                                    $notification->setDateTime(new DateTime());
                                    $this->_notificationManager->notify($notification);
                                }
                            }
                        }
                    }
                }
                $this->_log('Indexation process : End');

                //$dateTimeNow = new DateTime('now', $this->_timeZone);
                //print __METHOD__ . ' : job end_time=' . $dateTimeNow->getTimestamp() . "\n";
                fclose($logFile);
            }

            $dateTimeNow = new DateTime('now', $this->_timeZone);
            print __METHOD__ . ' : cron run end_time=' . $dateTimeNow->getTimestamp() . ", date=" . $dateTimeNow->format("d/m/Y H:i:s") . "\n";
        } catch (Exception $e) {
            print 'Exception ' . __METHOD__ . ' : error=' . $e->getMessage();
        }
	}
}
