<?php

namespace OCA\Dematpayslip\Db;

use JsonSerializable;

use OCP\AppFramework\Db\Entity;
use OCP\IConfig;
use OCP\IDBConnection;
use OCP\IL10N;
use OCP\Files\IRootFolder;

use OCP\Files\File;
//use OC\Files\Filesystem;

use DateTime;
use DateTimeZone;
use Exception;

use OCA\Pastell\Db\PastellConfigFlowMapper;
use OCA\Pastell\PastellAPI;


class DematpayslipArchive extends Entity implements JsonSerializable {

    /**
     * Const operation
     */
    const OPERATION_INITIALIZED = 0;
    const OPERATION_INDEXATION  = 1;
    const OPERATION_ARCHIVE     = 2;
    const OPERATION_FINISHED    = 3;

    /**
     * Const status
     */
    const STATUS_DELETED       = -1;
    const STATUS_INITIALIZED   = 0;
    const STATUS_INDEXED       = 1;
    const STATUS_CREATED       = 2;
    const STATUS_TITLED        = 3;
    const STATUS_FILE_SENT     = 4;
    const STATUS_SAE_SENT      = 5;
    const STATUS_METADATA_SENT = 6;
    const STATUS_ORIENTATION   = 7;

    /**
     * Const step status
     */
    const STEP_STATUS_TO_DO    = 0;
    const STEP_STATUS_PROGRESS = 1;
    const STEP_STATUS_FINISHED = 2;

    // messages
    private $_msgList = array(
        'error'   => array(),
        'success' => array(),
        'warning' => array(),
    );

    protected $idFile;
    protected $idUser;
    protected $filePath;
    protected $dematpayslipIdUser;
    protected $pastellIdDocument;
    protected $siret;
    protected $matricule;
    protected $year;
    protected $month;
    protected $lastname;
    protected $firstname;
    protected $email;
    protected $zipCity;
    protected $startDate;
    protected $endDate;
    protected $editionDate;
    protected $status;
    protected $statusTime;
    protected $pastellUrl;
    protected $pastellLastAction;
    protected $importTime;
    protected $checksum;
    protected $dematIdProcess;
    protected $filePagesNum;
    protected $operationAsk;
    protected $operationProgress;
    protected $error;

    public function jsonSerialize() {
        return [
            'id'                 => $this->id,
            'idFile'             => $this->idFile,
            'idUser'             => $this->idUser,
            'filePath'           => $this->filePath,
            'dematpayslipIdUser' => $this->dematpayslipIdUser,
            'pastellIdDocument'  => $this->pastellIdDocument,
            'siret'              => $this->siret,
            'matricule'          => $this->matricule,
            'year'               => $this->year,
            'month'              => $this->month,
            'lastname'           => $this->lastname,
            'firstname'          => $this->firstname,
            'email'              => $this->email,
            'zipCity'            => $this->zipCity,
            'startDate'          => $this->startDate,
            'endDate'            => $this->endDate,
            'editionDate'        => $this->editionDate,
            'status'             => $this->status,
            'status_time'        => $this->statusTime,
            'pastellUrl'         => $this->pastellUrl,
            'pastellLastAction'  => $this->pastellLastAction,
            'importTime'         => $this->importTime,
            'checksum'           => $this->checksum,
            'dematIdProcess'     => $this->dematIdProcess,
            'filePagesNum'       => $this->filePagesNum,
            'operationAsk'       => $this->operationAsk,
            'operationProgress'  => $this->operationProgress,
            'error'              => $this->error,
        ];
    }

    /**
     * Get message list
     *
     * @param   string  $type       [=''] Type of message (success, warning, error)
     * @return  array   Message list
     */
    public function getMsgList($type = '') {
        if (!empty($type) && isset($this->_msgList[$type])) {
            $msgList = $this->_msgList[$type];
        } else {
            $msgList = $this->_msgList;
        }

        return $msgList;
    }

    /**
     * Get operation label from id
     *
     * @param   string      $id_operation   Operation id
     * @return  string      Operation label
     */
    public static function getOperationLabelFrom($id_operation) {
        $label = '';

        $labelList = array(
            self::OPERATION_INITIALIZED  => 'Initialisation',
            self::OPERATION_INDEXATION => 'Indexation',
            self::OPERATION_ARCHIVE => 'Archivage',
            self::OPERATION_FINISHED  => 'Terminé',
        );

        if (isset($labelList[$id_operation])) {
            $label = $labelList[$id_operation];
        }

        return $label;
    }

    /**
     * Show operation ask label
     *
     * @return  string  Operation ask label
     */
    public function showOperationAskLabel() {
        $operationAskLabelList = array();

        $operationAskList = explode('+', $this->getOperationAsk());
        if (!empty($operationAskList)) {
            foreach ($operationAskList as $operationAsk) {
                $operationAskLabelList[] = self::getOperationLabelFrom($operationAsk);
            }
        }

        return implode(' | ', $operationAskLabelList);
    }

    /**
     * Show operation progress label
     *
     * @return  string  Operation progress label
     */
    public function showOperationProgressLabel() {
        return self::getOperationLabelFrom($this->getOperationProgress());
    }

    /**
     * Show status label
     *
     * @param   bool    $withHtml       [=false] With HTML tags
     * @return  string  Status label
     */
    public function showStatusLabel($withHtml = false) {
        $pastellLastAction = $this->getPastellLastAction();

        $statusLabelList = array(
            self::STATUS_DELETED       => 'Supprimé',
            self::STATUS_INITIALIZED   => 'Aucun',
            self::STATUS_INDEXED       => 'Indexé',
            self::STATUS_CREATED       => 'Créé',
            self::STATUS_TITLED        => 'Titre ajouté',
            self::STATUS_FILE_SENT     => 'Fichier envoyé',
            self::STATUS_SAE_SENT      => 'Envoyé au SAE',
            self::STATUS_METADATA_SENT => 'Méta-données transmises',
            self::STATUS_ORIENTATION   => 'Orienté',
        );

        $statusLabel  = $statusLabelList[$this->getStatus()];
        if (!empty($pastellLastAction)) {
            $statusLabel .= ' ';
            if ($withHtml) {
                if ($pastellLastAction == 'ar-recu-sae') {
                    $statusLabel .= '<span style="color: #777000;">';
                } else if ($pastellLastAction == 'accepter-sae') {
                    $statusLabel .= '<span style="color: #007700;">';
                } else if ($pastellLastAction == 'rejet-sae') {
                    $statusLabel .= '<span style="color: #770000;">';
                } else {
                    $statusLabel .= '<span>';
                }
            }
            $statusLabel .=  '[' . $pastellLastAction . ']';
            if ($withHtml) {
                $statusLabel .= '</span>';
            }
        }

        return $statusLabel;
    }

    /**
     * Save payslip file to index/archive from path and agent user
     *
     * @param   File                        $archiveFile                        Archive file
     * @param   DematpayslipArchiveMapper   $dematpayslipArchiveMapper          Archive mapper
     * @param   DematpayslipUser            $agentUser                          Agent user
     * @param   string                      $startDate                          Payslip start date
     * @param   string                      $endDate                            Payslip end date
     * @param   string                      $year                               Payslip year
     * @param   string                      $month                              Payslip month
     * @param   string                      $editionDate                        Payslip edition date
     * @param   int                         $status                             [=self::STATUS_INITIALIZED] Status
     * @param   DateTime                    $importDateTime                     [=null] Import date time
     * @param   int                         $idProcess                          [=0] Process id
     * @param   string                      $operationAsk                       [=''] Operation ask
     * @param   string                      $operationProgress                  [=0] Operation progress
     * @param   DateTimeZone                $timeZone                           [=null] Time zone
     *
     * @throws  Exception
     */
    public function saveFromFilePathAndAgent(File $archiveFile, DematpayslipArchiveMapper $dematpayslipArchiveMapper, DematpayslipUser $agentUser, $startDate, $endDate, $year, $month, $editionDate, $status = self::STATUS_INITIALIZED, DateTime $importDateTime = null, $idProcess = 0, $filePagesNum = '', $operationAsk = '', $operationProgress = 0, $timeZone = null) {
        $statusDateTime = new DateTime('now', $timeZone);
        //$archiveFile = Filesystem::getFileInfo($filePath);
        //$archiveFileChecksum = Filesystem::hash('md5', $filePath);
        $archiveFileChecksum = $archiveFile->hash('md5');

        $this->setIdFile($archiveFile->getId());
        $this->setFilePath($archiveFile->getPath());
        $this->setIdUser($agentUser->getUserId());
        $this->setDematpayslipIdUser($agentUser->getId());
        $this->setSiret($agentUser->getSiret());
        $this->setMatricule($agentUser->getMatricule());
        $this->setYear($year);
        $this->setMonth($month);
        $this->setLastname($agentUser->getLastname());
        $this->setFirstname($agentUser->getFirstname());
        $this->setEmail($agentUser->getEmail());
        $this->setZipCity($agentUser->getZipCity());
        $this->setStartDate($startDate);
        $this->setEndDate($endDate);
        $this->setEditionDate($editionDate);
        $this->setStatus($status);
        $this->setStatusTime($statusDateTime->getTimestamp());
        if ($importDateTime) {
            $this->setImportTime($importDateTime->getTimestamp());
        }
        $this->setChecksum($archiveFileChecksum);
        if ($idProcess > 0) {
            $this->setDematIdProcess($idProcess);
        }
        if (!empty($filePagesNum)) {
            $this->setFilePagesNum($filePagesNum);
        }
        if (!empty($operationAsk)) {
            $this->setOperationAsk($operationAsk);
        }
        if ($operationProgress > 0) {
            $this->setOperationProgress($operationProgress);
        }

        if (empty($this->getId())) {
            // insert new file to archive
            $dematpayslipArchiveMapper->insert($this);
        } else {
            // update file to archive
            $dematpayslipArchiveMapper->update($this);
        }
    }

    /**
     * Determine if it has already been archived (can't modify)
     *
     * @return  bool
     */
    public function isArchived() {
        $isArchived = false;

        if (!empty($this->getPastellIdDocument())) {
            if (in_array($this->getStatus(), [self::STATUS_ORIENTATION])) {
                $isArchived = true;
            }
        }

        return $isArchived;
    }

    /**
     * Archive process
     *
     * @param   string                      $appName                        App name
     * @param   IConfig                     $config                         Config
     * @param   IDBConnection               $dbConnection                   Database connection
     * @param   IL10N                       $l10n                           Lang
     * @param   IRootFolder                 $rootFolder                     Root folder
     * @param   DematpayslipArchiveMapper   $dematpayslipArchiveMapper      Archive mapper
     * @param   DematpayslipUserMapper      $dematpayslipUserMapper         User mapper
     * @param   DateTimeZone                $dateTimeZone                   [=null] Date time zone
     * @return  int                         <0 if errors else >0 on success
     */
    public function archiveProcess(string $appName, IConfig $config, IDBConnection $dbConnection, IL10N $l10n, IRootFolder $rootFolder, DematpayslipArchiveMapper $dematpayslipArchiveMapper, DematpayslipUserMapper $dematpayslipUserMapper, DateTimeZone $dateTimeZone = null) {
        $archiveFlowNum = 1; // archiving for 5 years by default
        // get dematpayslip user
        if (!empty($this->getDematpayslipIdUser())) {
            $dematpayslipUser = $dematpayslipUserMapper->find($this->getDematpayslipIdUser());
            if ($dematpayslipUser->getDisabled() == 0) {
                $archiveFlowNum = 2; // archiving for 50 years
            }
        }

        try {
            // Demat config
            $appDataDir = $config->getAppValue($appName, 'app_data_dir', 'dematpayslip');
            $userId = $config->getAppValue($appName, 'user_id', 'demat');

            // Pastell API config
            $pastellConfigFlowId = $config->getAppValue($appName, 'pastell_config_f' . $archiveFlowNum . '_id', '');
            $pastellConfigFlowMapper = new PastellConfigFlowMapper($dbConnection);
            $pastellConfigFlow = $pastellConfigFlowMapper->find($pastellConfigFlowId);

            // Pastell API connection
            $pastellAPI = new PastellAPI($pastellConfigFlow->getIdConfig());

            $canModifyArchive = false;
            if ($this->isArchived() === false) {
                $canModifyArchive = true;
            } else {
                // get and modify document last action
                $responseData = $pastellAPI->getEntityDocument($pastellConfigFlow->getIdE(), $this->getPastellIdDocument());

                $archiveLastAction = $responseData['last_action']['action'];
                if (!empty($archiveLastAction)) {
                    $this->setPastellLastAction($archiveLastAction);
                    $dematpayslipArchiveMapper->update($this);
                }
            }

            if ($canModifyArchive === false) {
                $this->_msgList['warning'][] = $l10n->t('L\'archive "' . $this->getFilePath() . '" a déjà été créée.');
            } else {
                $userFolder = $rootFolder->getUserFolder($userId);

                // 1. create document
                if (empty($this->getPastellIdDocument())) {
                    $nowDateTime = new DateTime('now', $dateTimeZone);
                    $responseData = $pastellAPI->postEntityDocument($pastellConfigFlow->getIdE(), array(
                        'type' => 'pdf-generique'
                    ));
                    $this->setPastellIdDocument($responseData['id_d']);
                    $this->setPastellLastAction($responseData['last_action']['action']);
                    $this->setStatus(self::STATUS_CREATED);
                    $this->setStatusTime($nowDateTime->getTimestamp());
                    $dematpayslipArchiveMapper->update($this);
                }
                // 2. set document title
                if ($this->getStatus() == self::STATUS_CREATED) {
                    $nowDateTime = new DateTime('now', $dateTimeZone);
                    $responseData = $pastellAPI->patchEntityDocument($pastellConfigFlow->getIdE(), $this->getPastellIdDocument(), array(
                        'libelle' => $this->getId() . '_' . $this->getSiret() . '_' . $this->getMatricule() . '_' . $this->getStartDate() . '_' . $this->getEndDate()
                    ));
                    $this->setPastellLastAction($responseData['content']['last_action']['action']);
                    $this->setStatus(self::STATUS_TITLED);
                    $this->setStatusTime($nowDateTime->getTimestamp());
                    $dematpayslipArchiveMapper->update($this);
                }
                // 3. send document file
                if ($this->getStatus() == self::STATUS_TITLED) {
                    $nowDateTime = new DateTime('now', $dateTimeZone);
                    $fileId = $this->getIdFile();
                    $filePath = $this->getFilePath();
                    $fileNode = null;
                    try {
                        $fileNode = $userFolder->getById($fileId)[0];
                    } catch (Exception $e) {
                        $this->_msgList['error'][] = $l10n->t('Impossible de récupérer le fichier a envoyer en archive "' . $filePath . '" [status=' . $this->getStatus() .'].');
                    }
                    if (!empty($fileNode) && $fileNode instanceof File) {
                        $fileName = $fileNode->getName();
                        $fileContent = $fileNode->getContent();
                        $responseData = $pastellAPI->postEntityDocumentFile($pastellConfigFlow->getIdE(), $this->getPastellIdDocument(), 'document', 0, $fileName, $fileContent);
                        $this->setPastellLastAction($responseData['content']['last_action']['action']);
                        $this->setStatus(self::STATUS_FILE_SENT);
                        $this->setStatusTime($nowDateTime->getTimestamp());
                        $dematpayslipArchiveMapper->update($this);
                    } else {
                        $this->_msgList['error'][] = $l10n->t('Mauvais fichier a envoyer en archive "' . $filePath . '" [status=' . $this->getStatus() .'].');
                    }
                }
                // 4. send to SAE
                if ($this->getStatus() == self::STATUS_FILE_SENT) {
                    $nowDateTime = new DateTime('now', $dateTimeZone);
                    $responseData = $pastellAPI->patchEntityDocument($pastellConfigFlow->getIdE(), $this->getPastellIdDocument(), array(
                        'envoi_sae' => true
                    ));
                    $this->setPastellLastAction($responseData['content']['last_action']['action']);
                    $this->setStatus(self::STATUS_SAE_SENT);
                    $this->setStatusTime($nowDateTime->getTimestamp());
                    $dematpayslipArchiveMapper->update($this);
                }
                // 5. create JSON file and add meta-data on document
                if ($this->getStatus() == self::STATUS_SAE_SENT) {
                    $saeConfigArr = array(
                        'matricule' => $this->getMatricule(),
                        'mois' => $this->getMonth(),
                        'annee' => $this->getYear(),
                        'nom' => $this->getLastname(),
                        'prenom' => $this->getFirstname(),
                        'date_edition' => $this->getEditionDate(),
                        'periode_2_de_paie' => $this->getEndDate(),
                        'periode_1_de_paie' => $this->getStartDate(),
                    );
                    $jsonFileName = 'sae_config.json';

                    //$appDataTmpPath = $userFolder->getPath() . DIRECTORY_SEPARATOR . $appDataDir . DIRECTORY_SEPARATOR . DematpayslipProcess::$appTmpDirName;
                    $appDataTmpPath = $appDataDir . DIRECTORY_SEPARATOR . DematpayslipProcess::$appTmpDirName;
                    $saeConfigFilePath = $appDataTmpPath . DIRECTORY_SEPARATOR . uniqid('sae');
                    if (!$saeConfigFilePath) {
                        $this->_msgList['error'][] = $l10n->t('Impossible de créer le nom du fichier temporaire "' . $appDataTmpPath . '" [status=' . $this->getStatus() .'].');
                    } else {
                        try {
                            //$tmpJsonFile = $rootFolder->newFile($saeConfigFilePath, json_encode($saeConfigArr));
                            $tmpJsonFile = $userFolder->newFile($saeConfigFilePath, json_encode($saeConfigArr));
                            $nowDateTime = new DateTime('now', $dateTimeZone);
                            $responseData = $pastellAPI->postEntityDocumentExternalData($pastellConfigFlow->getIdE(), $this->getPastellIdDocument(), 'sae_config', array(
                                'file_name' => $jsonFileName,
                                'file_content' => $tmpJsonFile->getContent()
                            ));
                            $this->setPastellLastAction($responseData['content']['last_action']['action']);
                            $this->setStatus(self::STATUS_METADATA_SENT);
                            $this->setStatusTime($nowDateTime->getTimestamp());
                            $dematpayslipArchiveMapper->update($this);
                            $tmpJsonFile->delete();
                        } catch (Exception $e) {
                            $this->_msgList['error'][] = $l10n->t('Impossible d\'écrire dans le fichier temporaire "' . $saeConfigFilePath . '" [status=' . $this->getStatus() .'].');
                        }
                    }
                }
                // 6. set orientation action on document
                if ($this->getStatus() == self::STATUS_METADATA_SENT) {
                    $nowDateTime = new DateTime('now', $dateTimeZone);
                    $responseData = $pastellAPI->postEntityDocumentAction($pastellConfigFlow->getIdE(), $this->getPastellIdDocument(), 'orientation');
                    $this->setStatus(self::STATUS_ORIENTATION);
                    $this->setStatusTime($nowDateTime->getTimestamp());
                    if ($responseData['result'] == true) {
                        // get and modify document last action
                        $responseData = $pastellAPI->getEntityDocument($pastellConfigFlow->getIdE(), $this->getPastellIdDocument());
                        if (!empty($responseData['last_action']['action'])) {
                            $this->setPastellLastAction($responseData['last_action']['action']);
                        }
                    }
                    $dematpayslipArchiveMapper->update($this);
                }

                if ($this->getStatus() < self::STATUS_ORIENTATION) {
                    $this->_msgList['error'][] = $l10n->t('L\'archive "' . $this->getFilePath() . '" n\'a pas été correctement créée [status=' . $this->getStatus() .'].');
                    $this->setError(1);
                } else {
                    $this->_msgList['success'][] = $l10n->t('L\'archive "' . $this->getFilePath() . '" a été créée.');
                    $this->setError(0);
                    $this->setOperationProgress(self::OPERATION_FINISHED);
                }
                $dematpayslipArchiveMapper->update($this);
            }

            if (count($this->_msgList['error']) > 0) {
                return -1;
            } else {
                return 1;
            }
        } catch (Exception $e) {
            $this->_msgList['error'][] = 'Exception ' . __METHOD__ . ' : error=' . $e->getMessage();
            return -1;
        }
    }
}
