<?php
namespace OCA\Dematpayslip\Db;

use JsonSerializable;

use OCP\AppFramework\Db\Entity;
use OCP\IConfig;
use OCP\IDateTimeZone;
use OCP\IDBConnection;
use OCP\IL10N;
use OCP\Share\IManager as IShareManager;
use OCP\Files\IRootFolder;
use OCP\IUserManager;

use OCP\Files\File;
use OC\Files\Filesystem;
use OCP\Files\Folder;

use DateTime;
use DateTimeZone;
use Exception;

// Pastell API
use OCA\Pastell\PastellAPI;
use OCA\Pastell\Db\PastellConfigFlowMapper;

class DematpayslipProcess extends Entity implements JsonSerializable {

    const LOG_SUCCESS = 0;
    const LOG_DEBUG   = 1;
    const LOG_WARNING = 2;
    const LOG_ERROR   = 3;

    const PROCESS_TIME_UNLOCK = 60; // seconds (max execution time)

    const PHASE_ID_UPLOAD = 0;
    const PHASE_ID_CHECK  = 1;
    const PHASE_ID_PRE_INDEXATION = 2;
    const PHASE_ID_INDEXATION = 3;

    const USER_REMOVE_UNKNOWN = 1; // remove unknown users from list (FIX SAML)

    /**
     * @var array Payslip PDF search field list
     */
    private $_payslipPDFSearchFieldList = array(
        'siret'     => array(3),
        'matricule' => array(10),
        'fullname'  => array(9),
        'zip_city'  => array(19, 20),
    );
//    private $_payslipPDFSearchFieldList = array(
//        'siret'     => array(4, 5),
//        'matricule' => array(20, 21),
//        'fullname'  => array(19, 20),
//        'zip_city'  => array(34, 35, 36, 37),
//    );

    /**
     * @var bool Fail on first error
     */
    private $_processFailOnFirstError = true;

    /**
     * @var string app name
     */
    private $_appName;

    /**
     * @var IConfig config
     */
    private $_config;

    /**
     * @var IDateTimeZone date time zone
     */
    private $_dateTimeZone;

    /**
     * @var IDBConnection database connection
     */
    private $_dbConnection;

    /**
     * @var IL10N lang
     */
    private $_l10n;

    /**
     * @var IRootFolder root folder
     */
    private $_rootFolder;

    /**
     * @var IShareManager share manager
     */
    private $_shareManager;

    /**
     * @var DateTimeZone time zone
     */
    private $_timeZone;

    /**
     * @var string user id
     */
    private $_userId;

    /**
     * @var IUserManager user manager
     */
    private $_userManager;

    /**
     * @var bool can cancel process
     */
    private $_canCancelProcess = false;

    /**
     * @var string CSV delimiter
     */
    private $_csvDelimiter = ';';

    /**
     * @var string CSV enclosure
     */
    private $_csvEnclosure = '"';

    /**
     * @var int CSV length
     */
    private $_csvLength = 0;

    /**
     * @var string CSV encode source
     */
    private $_csvEncodeFrom = 'UTF-8';

    /**
     * @var string CSV encode destination
     */
    private $_csvEncodeTo = 'UTF-8';

    /**
     * @var resource handler Log file
     */
    private $_logFile = null;

    /**
     * @var int log level
     */
    private $_logLevel = self::LOG_DEBUG;

    /**
     * @var string folder path
     */
    private $_uploadDirUserPath;
    private $_uploadDirFullPath;
    private $_checkDirUserPath;
    private $_checkDirFullPath;
    private $_reportDirUserPath;
    private $_reportDirFullPath;
    private $_exportDirUserPath;
    private $_exportDirFullPath;
    private $_tmpDirUserPath;
    private $_tmpDirFullPath;

    /**
     * @var string payslip no index dir
     */
    private $_payslipNoIndexDir = 'BulletinsNoIndex';

    /**
     * @var DematpayslipProcessMapper
     */
    private $_dematpayslipProcessMapper;

    /**
     * @var DematpayslipProcessSkipMapper
     */
    private $_dematpayslipProcessSkipMapper;

    /**
     * @var DematpayslipUserMapper
     */
    private $_dematpayslipUserMapper;

    /**
     * @var array[] dematpayslip config list
     */
    private $_dematpayslipConfigList = array();

    /**
     * @var array[] message list
     */
    private $_msgList = array(
        'error'   => array(),
        'success' => array(),
        'warning' => array(),
    );

    /**
     * Pastell app exists
     *
     * @var bool
     */
    private $_pastellAppExists = false;

    /**
     * @var string app tmp directory name
     */
    public static $appTmpDirName = 'tmp';

    protected $processing;
    protected $phase;
    protected $phaseStartTime;
    protected $phaseStep;
    protected $phaseStepNum;
    protected $mustTerminate;
    protected $hasError;
    protected $errorCode;
    protected $errorMsg;
    protected $pdfOrigin;
    protected $csvOrigin;
    protected $csvCheck;
    protected $userId;

    public function jsonSerialize() {
        return [
            'id'             => $this->id,
            'processing'     => $this->processing,
            'phase'          => $this->phase,
            'phaseStartTime' => $this->phaseStartTime,
            'phaseStep'      => $this->phaseStep,
            'phaseStepNum'   => $this->phaseStepNum,
            'mustTerminate'  => $this->mustTerminate,
            'hasError'       => $this->hasError,
            'errorCode'      => $this->errorCode,
            'errorMsg'       => $this->errorMsg,
            'csvOrigin'      => $this->csvOrigin,
            'pdfOrigin'      => $this->pdfOrigin,
            'csvCheck'       => $this->csvCheck,
            'userId'         => $this->userId,
        ];
    }

    /**
     * Check if payslip file has not been archived yet
     *
     * @param   DematpayslipArchiveMapper   $dematpayslipArchiveMapper      Archive table mapper
     * @param   DematpayslipUser            $agentUser                      Agent user
     * @param   string                      $startDate                      Period start date
     * @param   string                      $endDate                        Period end date
     * @return  mixed|DematpayslipArchive|\OCP\AppFramework\Db\Entity|null
     *
     * @throws  Exception
     */
    private function _archiveCheckBeforeSave(DematpayslipArchiveMapper $dematpayslipArchiveMapper, DematpayslipUser $agentUser, $startDate, $endDate) {
        $dematpayslipArchive = null;
        $dematpayslipArchiveExistList = $dematpayslipArchiveMapper->findAllByAgentAndPeriod($agentUser->getSiret(), $agentUser->getMatricule(), $agentUser->getLastname(), $agentUser->getFirstname(), $startDate, $endDate);

        if (empty($dematpayslipArchiveExistList)) {
            $dematpayslipArchive = new DematpayslipArchive();
        } elseif (count($dematpayslipArchiveExistList) === 1) {
            if (!empty($dematpayslipArchiveExistList[0]->getPastellIdDocument())) {
                $warningMsg = $this->_l10n->t('Le bulletin de paie de l\'agent "' . $agentUser->getLastname() . ' ' . $agentUser->getFirstname() . '" pour la période du ' . $startDate . ' au ' . $endDate . ' est déjà en cours d\'archivage.');
                $this->_logCsvLine($warningMsg, self::LOG_WARNING, $agentUser, $startDate, $endDate);
                $this->_processWarningMsg($warningMsg);
            } else {
                $dematpayslipArchive = $dematpayslipArchiveExistList[0];
            }
        } else {
            $warningMsg = $this->_l10n->t('Plusieurs bulletins de paie de l\'agent "' . $agentUser->getLastname() . ' ' . $agentUser->getFirstname() . '" pour la période du ' . $startDate . ' au ' . $endDate . ' existent déjà.');
            $this->_logCsvLine($warningMsg, self::LOG_WARNING, $agentUser, $startDate, $endDate);
            $this->_processWarningMsg($warningMsg);
        }

        return $dematpayslipArchive;
    }

    /**
     * Check Pastell app exists
     */
    private function _checkPastellAppExists() {
        $pastellAppExists = false;

        $pastellAppDir = __DIR__ . '/../../../pastell';
        if (is_dir($pastellAppDir)) $pastellAppExists = true;

        return $pastellAppExists;
    }

    /**
     * Find CSV and PDF files in uploaded directory
     *
     * @return  array   CSV and PDF filenames
     */
    private function _findUploadedCSVAndPDF() {
        $uploadFileCSVName = '';
        $uploadFilePDFName = '';
        $uploadDirFullPath = $this->_uploadDirFullPath;
        if (file_exists($uploadDirFullPath)) {
            $scanDir = scandir($uploadDirFullPath);
            if (!empty($scanDir)) {
                foreach ($scanDir as $fileName) {
                    if (!in_array($fileName, array('.', '..'))) {
                        $uploadFilePath = $uploadDirFullPath . DIRECTORY_SEPARATOR . $fileName;
                        if (is_file($uploadFilePath)) {
                            $uploadFileType = mime_content_type($uploadFilePath);
                            if ($uploadFileType == 'application/pdf') {
                                $uploadFilePDFName = $fileName;
                            } else if ($uploadFileType == 'text/plain') {
                                $uploadFileCSVName = $fileName;
                            }
                        }
                    }

                    if (!empty($uploadFileCSVName) && !empty($uploadFilePDFName)) {
                        break;
                    }
                }
            }
        }

        return array('csv' => $uploadFileCSVName, 'pdf' => $uploadFilePDFName);
    }

    /**
     * Find all files (recursive)
     *
     * @param   string      $filePath       File path
     * @return  array
     */
    private function _findAllFiles($filePath, &$fileList = array()) {
        $isDir = Filesystem::is_dir($filePath);

        if ($isDir) {
            $payslipPDFFileList = Filesystem::getDirectoryContent($filePath);

            foreach ($payslipPDFFileList as $payslipPDFFile) {
                $newFilePath = $filePath . DIRECTORY_SEPARATOR . $payslipPDFFile->getName();
                $isFile = Filesystem::is_file($newFilePath);
                if ($isFile) {
                    $fileList[] = $payslipPDFFile;
                } else {
                    $this->_findAllFiles($newFilePath, $fileList);
                }
            }
        }

        return $fileList;
    }

    /**
     * Remove all files in a directory
     *
     * @param   string      $dirPath    Directory path
     */
    private static function _removeAllFilesDir($dirPath) {
        $scanDir = scandir($dirPath);
        if (!empty($scanDir)) {
            foreach ($scanDir as $fileName) {
                if (!in_array($fileName, array('.', '..'))) {
                    $filePath = $dirPath . DIRECTORY_SEPARATOR . $fileName;
                    if (is_file($filePath)) {
                        unlink($filePath);
                    }
                }
            }
        }
    }

    /**
     * Remove all unknown users (FIX for multiple users with SAML)
     *
     * @param   array   $userFindByEmailList        User list
     * @return  array   Users list without unknown
     */
    private static function _removeUnknownUser($userFindByEmailList) {
        $userList = $userFindByEmailList;

        foreach ($userList as $key => $user) {
            if (preg_match('#^[0-9A-F]{8}\-[0-9A-F]{4}\-[0-9A-F]{4}\-[0-9A-F]{4}\-[0-9A-F]{12}#', $user->getDisplayName())) {
                unset($userList[$key]);
            }
        }

        return $userList;
    }

    /**
     * Cancel process
     */
    private function _processCancel() {
        if ($this->getMustTerminate() <= 0) {
            // clean upload directory
            if (is_dir($this->_uploadDirFullPath)) {
                self::_removeAllFilesDir($this->_uploadDirFullPath);
            }

            // clean check directory
            if (is_dir($this->_checkDirFullPath)) {
                self::_removeAllFilesDir($this->_checkDirFullPath);
            }

            // init process and delete archives links (status initialized and process must terminate = 0)
            $this->_processInit();
            $this->_dematpayslipProcessMapper->update($this);
        }
    }

    /**
     * Set process error message
     *
     * @param   int         $errorCode      Error code
     * @param   string      $errorMsg       Error message
     * @param   string      $phaseStep      [=''] Phase step
     * @param   int         $phaseStepNum   [=0] Phase step num
     */
    private function _processErrorMsg($errorCode, $errorMsg, $phaseStep = '', $phaseStepNum = 0) {
        $this->setHasError(1);
        $this->_msgList['error'][] = $errorMsg;
        $this->setErrorCode($errorCode);
        $this->setErrorMsg(implode('<br />', $this->_msgList['error']));
        $this->setPhaseStep($phaseStep);
        $this->setPhaseStepNum(intval($phaseStepNum));
        $this->_dematpayslipProcessMapper->update($this);
    }

    /**
     * Set process warning message
     *
     * @param   string      $warningMsg       Warning message
     */
    private function _processWarningMsg($warningMsg) {
        $this->_msgList['warning'][] = $warningMsg;
    }

    /**
     * Csv log file
     *
     * @param   array   $colList    Column name list
     */
    private function _logCsvCols($colList) {
        if ($this->_logFile !== null) {
            fputcsv($this->_logFile, $colList, $this->_csvDelimiter, $this->_csvEnclosure);
        }
    }

    /**
     * Log csv line
     *
     * @param   string                  $msg            Log message
     * @param   int                     $level          [=LOG_DEBUG] Log level
     * @param   DematpayslipUser        $agentUser      [=null] Agent user
     * @param   string                  $startDate      [=''] Start date
     * @param   string                  $endDate        [=''] End date
     *
     * @throws  Exception
     */
    private function _logCsvLine($msg, $level = self::LOG_DEBUG, DematpayslipUser $agentUser = null, $startDate = '', $endDate = '') {
        if ($this->_logFile !== null) {
            if ($this->_logLevel <= $level || $level === self::LOG_SUCCESS) {
                $dateTimeNow = new DateTime('now', $this->_timeZone);
                $logLevel = '';
                switch ($level) {
                    case self::LOG_SUCCESS :
                        $logLevel = 'SUCCESS';
                        break;

                    case self::LOG_DEBUG :
                        $logLevel = 'DEBUG';
                        break;

                    case self::LOG_WARNING :
                        $logLevel = 'WARNING';
                        break;

                    case self::LOG_ERROR :
                        $logLevel = 'ERROR';
                        break;

                    default : break;
                }

                $csvLine = array(
                    $dateTimeNow->format('Y-m-d'),
                    $dateTimeNow->format('H:i:s'),
                    $logLevel,
                    $msg,
                    (!empty($agentUser) ? $agentUser->getSiret() : ''),
                    (!empty($agentUser) ? $agentUser->getMatricule() : ''),
                    (!empty($agentUser) ? $agentUser->getLastname() . ' ' . $agentUser->getFirstname() : ''),
                    (!empty($agentUser) ? $agentUser->getZipCity() : ''),
                    $startDate,
                    $endDate,
                    (!empty($agentUser) ? $agentUser->getEmail() : ''),
                );
                fputcsv($this->_logFile, $csvLine, $this->_csvDelimiter, $this->_csvEnclosure);
            }
        }
    }

    /**
     * Init process
     */
    private function _processInit() {
        $this->setPhase(0);
        $this->setPhaseStartTime(0);
        $this->setPhaseStep('');
        $this->setPhaseStepNum(0);
        $this->setMustTerminate(0);
        $this->setHasError(0);
        $this->setErrorCode(0);
        $this->setErrorMsg('');
        $this->setCsvOrigin('');
        $this->setPdfOrigin('');
        $this->setCsvCheck('');
        $this->setUserId('');
    }

    /**
     * Init application folders
     *
     * @throws Exception
     */
    private function _initDematFolders() {
        $appDataDirPath           = (string) $this->_dematpayslipConfigList['app_data_dir']; // app directory of operator user (upload, check, report, ...)
        $this->_checkDirUserPath  = $appDataDirPath . DIRECTORY_SEPARATOR . 'check';
        $this->_reportDirUserPath = $appDataDirPath . DIRECTORY_SEPARATOR . 'report';
        $this->_uploadDirUserPath = $appDataDirPath . DIRECTORY_SEPARATOR . 'upload';
        $this->_exportDirUserPath = $appDataDirPath . DIRECTORY_SEPARATOR . 'export';
        $this->_tmpDirUserPath    = $appDataDirPath . DIRECTORY_SEPARATOR . self::$appTmpDirName;

        // get user for payslip
        $user = $this->_userManager->get($this->_userId);
        if (!$user) {
            $this->_msgList['error'][] = $this->_l10n->t('L\'utilisateur "' . $this->_userId . '" n\'existe pas.');
        }

        // create application user directory
        if (!$this->hasError()) {
            $userHomeAppFilesFullPath = $user->getHome() . DIRECTORY_SEPARATOR . 'files';
            $this->_checkDirFullPath  = $userHomeAppFilesFullPath . DIRECTORY_SEPARATOR . $this->_checkDirUserPath;
            $this->_reportDirFullPath = $userHomeAppFilesFullPath . DIRECTORY_SEPARATOR . $this->_reportDirUserPath;
            $this->_uploadDirFullPath = $userHomeAppFilesFullPath . DIRECTORY_SEPARATOR . $this->_uploadDirUserPath;
            $this->_exportDirFullPath = $userHomeAppFilesFullPath . DIRECTORY_SEPARATOR . $this->_exportDirUserPath;
            $this->_tmpDirFullPath    = $userHomeAppFilesFullPath . DIRECTORY_SEPARATOR . $this->_tmpDirUserPath;

            // mount user directory
            $userFolder = $this->_rootFolder->getUserFolder($this->_userId);

            // create payslip directory
            $payslipDir = (string) $this->_dematpayslipConfigList['payslip_dir'];
            try {
                $userFolder->get($payslipDir);
            } catch (Exception $e) {
                $createdFolder = $userFolder->newFolder($payslipDir);
                if (!$createdFolder) {
                    $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier des bulletins "' . $payslipDir . '".');
                }
            }

            // create payslip not to be indexed folder
            try {
                $userFolder->get($this->_payslipNoIndexDir);
            } catch (Exception $e) {
                $createdFolder = $userFolder->newFolder($this->_payslipNoIndexDir);
                if (!$createdFolder) {
                    $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier des bulletins non indexes "' . $this->_payslipNoIndexDir . '".');
                }
            }

            // payslip application directory
            try {
                $userFolder->get($appDataDirPath);
            } catch (Exception $e) {
                $createdFolder = $userFolder->newFolder($appDataDirPath);
                if (!$createdFolder) {
                    $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier de l\'application "' . $appDataDirPath . '".');
                }
            }

            if (!$this->hasError()) {
                // check directory
                try {
                    $userFolder->get($this->_checkDirUserPath);
                } catch (Exception $e) {
                    $createdFolder = $userFolder->newFolder($this->_checkDirUserPath);
                    if (!$createdFolder) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier de l\'application "' . $this->_checkDirUserPath . '".');
                    }
                }

                // report directory
                try {
                    $userFolder->get($this->_reportDirUserPath);
                } catch (Exception $e) {
                    $createdFolder = $userFolder->newFolder($this->_reportDirUserPath);
                    if (!$createdFolder) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier de l\'application "' . $this->_reportDirUserPath . '".');
                    }
                }

                // upload directory
                try {
                    $userFolder->get($this->_uploadDirUserPath);
                } catch (Exception $e) {
                    $createdFolder = $userFolder->newFolder($this->_uploadDirUserPath);
                    if (!$createdFolder) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier de l\'application "' . $this->_uploadDirUserPath . '".');
                    }
                }

                // export directory
                try {
                    $userFolder->get($this->_exportDirUserPath);
                } catch (Exception $e) {
                    $createdFolder = $userFolder->newFolder($this->_exportDirUserPath);
                    if (!$createdFolder) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de créer le dossier de l\'application "' . $this->_exportDirUserPath . '".');
                    }
                }

                // tmp directory
                try {
                    $userFolder->get($this->_tmpDirUserPath);
                } catch (Exception $e) {
                    $createdFolder = $userFolder->newFolder($this->_tmpDirUserPath);
                    if (!$createdFolder) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de créeer le dossier de l\'application "' . $this->_tmpDirUserPath . '".');
                    }
                }
            }
        }
    }

    /**
     * Init for agent user
     *
     * @throws  Exception
     */
    private function _initDematUser() {
        // init
        $appPath = \OC::$server->getAppManager()->getAppPath($this->_appName);

        // depends
        require_once $appPath . '/vendor/autoload.php';

        $this->_initDematFolders();
    }

    /**
     * Get phase name
     *
     * @param   int         $phaseId    Phase id
     * @return  string      Phase name
     */
    private static function _getPhaseName($phaseId) {
        $phaseName = '';

        switch ($phaseId) {
            case self::PHASE_ID_UPLOAD :
                $phaseName = 'Upload';
                break;

            case self::PHASE_ID_CHECK :
                $phaseName = 'Vérification';
                break;

            case self::PHASE_ID_PRE_INDEXATION :
                $phaseName = 'Pré-indexation';
                break;

            case self::PHASE_ID_INDEXATION :
                $phaseName = 'Indexation';
                break;

            default : break;
        }

        return $phaseName;
    }

    /**
     * Check if process is locked (processing or locked by user)
     */
    private function _processCheck() {
        // automatic unlock process (ex : max execution time)
        if ((time() - $this->getPhaseStartTime()) > self::PROCESS_TIME_UNLOCK) {
            $this->setProcessing(0);
            $this->_dematpayslipProcessMapper->update($this);
        }

        $this->_canCancelProcess = false;
        // check if process was launched by the same user
        if (!empty($this->getUserId()) && $this->_userId===$this->getUserId()) {
            if ($this->getProcessing()==0 && $this->getMustTerminate()==0) {
                $this->_canCancelProcess = true;
            }
        } else {
            // time too long and greater than automatic unlock process time
            if ((time() - $this->getPhaseStartTime()) > self::PROCESS_TIME_UNLOCK) {
                if ($this->getProcessing()==0 && $this->getMustTerminate()==0) {
                    $this->_canCancelProcess = true;
                }
            }
        }

        if ($this->getProcessing()>0) {
            $this->_msgList['error'][] = $this->_l10n->t('Une opération "' . self::_getPhaseName($this->getPhase()). '" est déjà en cours...');

            if (!empty($this->getUserId()) && $this->_userId!==$this->getUserId()) {
                $this->_msgList['error'][] = $this->_l10n->t('Opération démarrée par l\'utilisateur "' . $this->getUserId() . '".');
            } else {
                $this->_msgList['error'][] = $this->_l10n->t('Veuillez patienter');
            }
        } else if (!empty($this->getUserId()) && $this->_userId!==$this->getUserId()) {
            $this->_msgList['error'][] = $this->_l10n->t('Une opération "' . self::_getPhaseName($this->getPhase()) . '" a déjà été démarrée par l\'utilisateur "' . $this->getUserId() . '".');
        } else if ($this->getHasError()>0) {
            if ($this->getMustTerminate()==1) {
                // last process terminate with errors, we have to determine action to unlock it
                $this->_msgList['error'][] = $this->_l10n->t('L\'opération "' . self::_getPhaseName($this->getPhase()) . '" s\'est terminée avec les erreurs suivantes') . ' : ';
                $this->_msgList['error'][] = $this->getErrorMsg();
            }
        }
    }

    /**
     * Get archive list
     *
     * @param   int         $stepStatus             [=-1] for all (not deleted), 0 for step init, 1 for step in progress and 2 for step archived
     * @param   bool        $refresh                [=false] not to refresh list form API, else true
     * @return  array       Archive list
     */
    public function getArchiveList($stepStatus = -1, $refresh = false) {
        // get archive list
        $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

        if ($refresh === true) {
            if ($stepStatus === -1) {
                $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAll(0);
            } else {
                $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAllByStepStatus($stepStatus);
            }

            if (!empty($dematpayslipArchiveList)) {
                foreach ($dematpayslipArchiveList as $dematpayslipArchive) {
                    if (empty($dematpayslipArchive->getPastellIdDocument()))  continue;

                    $archiveFlowNum = 1; // archiving for 5 years by default

                    // get dematpayslip user
                    if (!empty($dematpayslipArchive->getDematpayslipIdUser())) {
                        $dematpayslipUser = $this->_dematpayslipUserMapper->find($dematpayslipArchive->getDematpayslipIdUser());
                        if ($dematpayslipUser->getDisabled() == 0) {
                            $archiveFlowNum = 2; // archiving for 50 years
                        }
                    }

                    // Pastell API config
                    $pastellConfigFlowId = $this->_config->getAppValue($this->_appName, 'pastell_config_f' . $archiveFlowNum . '_id', '');
                    $pastellConfigFlowMapper = new PastellConfigFlowMapper($this->_dbConnection);
                    $pastellConfigFlow = $pastellConfigFlowMapper->find($pastellConfigFlowId);

                    try {
                        // Pastell API connection
                        $pastellAPI = new PastellAPI($pastellConfigFlow->getIdConfig());

                        // get and modify document last action
                        $responseData = $pastellAPI->getEntityDocument($pastellConfigFlow->getIdE(), $dematpayslipArchive->getPastellIdDocument());
                        if (!empty($responseData['last_action']['action'])) {
                            $dematpayslipArchive->setPastellLastAction($responseData['last_action']['action']);
                            $dematpayslipArchiveMapper->update($dematpayslipArchive);
                        }
                    } catch (Exception $e) {
                        $this->_msgList['error'][] = $e->getMessage();
                    }

                    if ($this->hasError())    break;
                }
            }

            if (!$this->hasError()) {
                $this->_msgList['success'][] = $this->_l10n->t('Liste actualisée avec succès.');
            }
        }

        if ($stepStatus === -1) {
            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAll(0);
        } else {
            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAllByStepStatus($stepStatus);
        }

        return $dematpayslipArchiveList;
    }

    /**
     * Get can cancel process
     *
     * @return bool
     */
    public function getCanCancelProcess() {
        return $this->_canCancelProcess;
    }

    /**
     * Get export dir user path
     *
     * @return mixed
     */
    public function getExportDirUserPath() {
        return $this->_exportDirUserPath;
    }

    /**
     * Get message list
     *
     * @param   string  $type       [=''] Type of message (success, warning, error)
     * @return  array   Message list
     */
    public function getMsgList($type = '') {
        if (!empty($type) && isset($this->_msgList[$type])) {
            $msgList = $this->_msgList[$type];
        } else {
            $msgList = $this->_msgList;
        }

        return $msgList;
    }

    /**
     * Get Pastell app exists
     *
     * @return bool
     */
    public function getPastellAppExists() {
        return $this->_pastellAppExists;
    }

    /**
     * Determine if process contains errors
     *
     * @return bool
     */
    public function hasError() {
        if (count($this->_msgList['error']) <= 0) {
            return false;
        } else {
            return true;
        }
    }

    /**
     * Load
     *
     * @param   string                          $appName                            App name
     * @param   IConfig                         $config                             Config
     * @param   IDBConnection                   $dbConnection                       Database connection
     * @param   IL10N                           $l10n                               Lang
     * @param   IRootFolder                     $rootFolder                         Root folder
     * @param   IShareManager                   $shareManager                       Share manager
     * @param   IUserManager                    $userManager                        User manager
     * @param   string                          $userId                             User id
     * @param   DematpayslipProcessMapper       $dematpayslipProcessMapper          Process mapper
     * @param   DematpayslipProcessSkipMapper   $dematpayslipProcessSkipMapper      Process skip mapper
     * @param   DematpayslipUserMapper          $dematpayslipUserMapper             User mapper
     * @param   DateTimeZone                    $timeZone                           [=null] Time zone
     * @return  array[]                         Config list
     */
    public function load(string $appName, IConfig $config, IDBConnection $dbConnection, IL10N $l10n, IRootFolder $rootFolder, IShareManager $shareManager, IUserManager $userManager, string $userId, DematpayslipProcessMapper $dematpayslipProcessMapper, DematpayslipProcessSkipMapper $dematpayslipProcessSkipMapper, DematpayslipUserMapper $dematpayslipUserMapper, DateTimeZone $timeZone = null) {
        $this->_appName = $appName;
        $this->_config = $config;
        //$this->_dateTimeZone = $dateTimeZone;
        $this->_dbConnection = $dbConnection;
        $this->_l10n = $l10n;
        $this->_rootFolder = $rootFolder;
        $this->_shareManager = $shareManager;
        $this->_timeZone = $timeZone;
        $this->_userManager = $userManager;
        $this->_userId = $userId;

        $this->_dematpayslipProcessMapper = $dematpayslipProcessMapper;
        $this->_dematpayslipProcessSkipMapper = $dematpayslipProcessSkipMapper;
        $this->_dematpayslipUserMapper = $dematpayslipUserMapper;

        $this->_pastellAppExists = $this->_checkPastellAppExists();
        $this->_loadConfig();

        return $this->_dematpayslipConfigList;
    }

    /**
     * Load config
     */
    private function _loadConfig() {
        $this->_dematpayslipConfigList = self::loadConfigFrom($this->_appName, $this->_config, $this->_l10n,$this->_msgList);
    }

    /**
     * Load config from parameters
     *
     * @param   string  $appName    App name
     * @param   IConfig $config     Config
     * @param   IL10N   $l10n       Lang
     * @param   array   $msgList    Message list
     * @return  array
     */
    public static function loadConfigFrom(string $appName, IConfig $config, IL10N $l10n, &$msgList) {
        $configList = array();

        $configList['payslip_dir']  = $config->getAppValue($appName, 'payslip_dir', 'Bulletins');       // Bulletins
        $configList['user_id']      = $config->getAppValue($appName, 'user_id', 'demat');               // demat
        $configList['app_data_dir'] = $config->getAppValue($appName, 'app_data_dir', 'dematpayslip');   // dematpayslip directory of user demat (upload, check, report, ...)
        if (empty($configList['payslip_dir']) || empty($configList['user_id'])) {
            $msgList['error'][] = $l10n->t('Veuillez vérifier la configuration de cette application.');
        }

        $configList['pastell_use'] = $config->getAppValue($appName, 'pastell_use', '');
        if (!empty($configList['pastell_use'])) {
            // check Pastell API config
            $configList['pastell_config_f1_id'] = $config->getAppValue($appName, 'pastell_config_f1_id', '');
            $configList['pastell_config_f2_id'] = $config->getAppValue($appName, 'pastell_config_f2_id', '');
            if (!($configList['pastell_config_f1_id'] > 0) || !($configList['pastell_config_f2_id'] > 0)) {
                $msgList['error'][] = $l10n->t('Veuillez vérifier la configuration de cette application.');
            }
        }

        return $configList;
    }

    /**
     * Cancel a process
     *
     * @throws  Exception
     */
    public function cancel() {
        $this->_initDematUser();

        $this->_processCheck();
        if ($this->_canCancelProcess === true) {
            $this->_processCancel();
        }
    }

    /**
     * Index
     *
     * @throws Exception
     */
    public function index() {
        $this->_initDematUser();

        $this->_processCheck();
        if (!$this->hasError()) {
            if ($this->getPhase()==self::PHASE_ID_UPLOAD) {
                // scan uploaded files
                $uploadedCSVAndPDFArr = $this->_findUploadedCSVAndPDF();
                $uploadFileCSVName = $uploadedCSVAndPDFArr['csv'];
                $uploadFilePDFName = $uploadedCSVAndPDFArr['pdf'];

                if (!empty($uploadFileCSVName) && !empty($uploadFilePDFName)) {
                    $this->setPhase(self::PHASE_ID_CHECK);
                    $this->setCsvOrigin($uploadFileCSVName);
                    $this->setPdfOrigin($uploadFilePDFName);
                    $this->_dematpayslipProcessMapper->update($this);
                }
            }
        }
    }

    /**
     * Upload
     *
     * @throws Exception
     */
    public function upload() {
        $this->_initDematUser();

        $this->_processCheck();
        if (!$this->hasError()) {
            $dateTimeNow = new DateTime('now', $this->_timeZone);

            $this->setProcessing(1);
            $this->setPhase(self::PHASE_ID_UPLOAD);
            $this->setPhaseStartTime($dateTimeNow->getTimestamp());
            $this->setPhaseStepNum(0);
            $this->setHasError(0);
            $this->setErrorCode(0);
            $this->setErrorMsg('');
            $this->setMustTerminate(0);
            $this->setCsvOrigin('');
            $this->setPdfOrigin('');
            $this->_dematpayslipProcessMapper->update($this);

            // check upload CSV
            $uploadFileCSVName    = '';
            $uploadFileCSVTmpName = '';
            if (!isset($_FILES['payslips_csv_file'])) {
                $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier CSV.');
                $this->_processErrorMsg(-1, $errorMsg);
            } else {
                $uploadFileArr = $_FILES['payslips_csv_file'];
                //var_dump($uploadFileArr['type']);
                if (empty($uploadFileArr['name'])) {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier CSV.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if (!in_array($uploadFileArr['type'], array('text/csv', 'application/octet-stream', 'application/vnd.ms-excel'))) {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier CSV valide.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if ($uploadFileArr['size']<=0) {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier CSV non vide.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if (empty($uploadFileArr['tmp_name'])) {
                    $errorMsg = $this->_l10n->t('Impossible d\'uploader ce fichier CSV.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if ($uploadFileArr['error']!==UPLOAD_ERR_OK) {
                    $errorMsg = $this->_l10n->t('Echec d\'upload du fichier CSV [=' . $uploadFileArr['error'] . '].');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else {
                    // upload success
                    $uploadFileCSVName    = $uploadFileArr['name'];
                    $uploadFileCSVTmpName = $uploadFileArr['tmp_name'];
                }
            }

            // check upload PDF
            $uploadFilePDFName    = '';
            $uploadFilePDFTmpName = '';
            if (!isset($_FILES['payslips_pdf_file'])) {
                $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier PDF.');
                $this->_processErrorMsg(-1, $errorMsg);
            } else {
                $uploadFileArr = $_FILES['payslips_pdf_file'];
                if (empty($uploadFileArr['name'])) {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier PDF.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if ($uploadFileArr['type']!='application/pdf') {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier PDF valide.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if ($uploadFileArr['size']<=0) {
                    $errorMsg = $this->_l10n->t('Veuillez sélectionner un fichier PDF non vide.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if (empty($uploadFileArr['tmp_name'])) {
                    $errorMsg = $this->_l10n->t('Impossible d\'uploader ce fichier PDF.');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else if ($uploadFileArr['error']!==UPLOAD_ERR_OK) {
                    $errorMsg = $this->_l10n->t('Echec d\'upload du fichier PDF [=' . $uploadFileArr['error'] . '].');
                    $this->_processErrorMsg(-1, $errorMsg);
                } else {
                    // upload success
                    $uploadFilePDFName    = $uploadFileArr['name'];
                    $uploadFilePDFTmpName = $uploadFileArr['tmp_name'];
                }
            }

            if ($this->getHasError()<=0) {
                // clean upload directory
                if (is_dir($this->_uploadDirFullPath)) {
                    self::_removeAllFilesDir($this->_uploadDirFullPath);
                }

                // upload CSV
                $uploadFilePath = $this->_uploadDirFullPath . DIRECTORY_SEPARATOR . $uploadFileCSVName;
                if (!move_uploaded_file($uploadFileCSVTmpName, $uploadFilePath)) {
                    $errorMsg = $this->_l10n->t('Echec d\'upload du fichier CSV dans le répertoire ' . $uploadFilePath . '.');
                    $this->_processErrorMsg(-1, $errorMsg);
                }

                // upload PDF
                $uploadFilePath = $this->_uploadDirFullPath . DIRECTORY_SEPARATOR . $uploadFilePDFName;
                if (!move_uploaded_file($uploadFilePDFTmpName, $uploadFilePath)) {
                    $errorMsg = $this->_l10n->t('Echec d\'upload du fichier PDF dans le répertoire ' . $uploadFilePath . '.');
                    $this->_processErrorMsg(-1,$errorMsg);
                }
            }

            if ($this->getHasError()>0) {
                if ($this->getPhase()==self::PHASE_ID_UPLOAD) {
                    // scan uploaded files
                    $uploadedCSVAndPDFArr = $this->_findUploadedCSVAndPDF();
                    $uploadFileCSVName = $uploadedCSVAndPDFArr['csv'];
                    $uploadFilePDFName = $uploadedCSVAndPDFArr['pdf'];

                    if (!empty($uploadFileCSVName) && !empty($uploadFilePDFName)) {
                        $this->setPhase(self::PHASE_ID_CHECK);
                        $this->setCsvOrigin($uploadFileCSVName);
                        $this->setPdfOrigin($uploadFilePDFName);
                        $this->setUserId($this->_userId);
                        $this->_dematpayslipProcessMapper->update($this);
                    }
                }
            } else {
                $this->_msgList['success'][] = $this->_l10n->t('Opération "' . self::_getPhaseName($this->getPhase()) . '" terminée avec succès.');
                $this->setPhase(self::PHASE_ID_CHECK);
                $this->setCsvOrigin($uploadFileCSVName);
                $this->setPdfOrigin($uploadFilePDFName);
                $this->setUserId($this->_userId);
            }

            $this->setProcessing(0);
            $this->_dematpayslipProcessMapper->update($this);
        }
    }

    /**
     * Check
     *
     * @throws Exception
     */
    public function check() {
        $this->_initDematUser();
        $this->_processCheck();
        if ($this->getPhase() != self::PHASE_ID_CHECK) {
            $this->_msgList['error'][] = $this->_l10n->t('Opération impossible : phase en cours "' . self::_getPhaseName($this->getPhase()) . '".');
        }

        if (!$this->hasError()) {
            $this->_processFailOnFirstError = false;
            $dateTimeNow = new DateTime('now', $this->_timeZone);

            $this->setProcessing(1);
            $this->setPhase(self::PHASE_ID_CHECK);
            $this->setPhaseStartTime($dateTimeNow->getTimestamp());
            $this->setPhaseStepNum(0);
            $this->setMustTerminate(0);
            $this->setHasError(0);
            $this->setErrorCode(0);
            $this->setErrorMsg('');
            $this->setUserId($this->_userId);
            $this->_dematpayslipProcessMapper->update($this);

            // empty process skip lines
            $this->_dematpayslipProcessSkipMapper->deleteAllByIdProcess($this->getId());

            // clean "check" directory
            if (is_dir($this->_checkDirFullPath)) {
                self::_removeAllFilesDir($this->_checkDirFullPath);
            }

            $uploadFilePDFName = $this->getPdfOrigin();
            $uploadFileCSVName = $this->getCsvOrigin();

            // log file
            $userFolder = $this->_rootFolder->getUserFolder($this->_userId);
            $logFileName = $this->getPhase() . '_log';
            $logFilePath = $this->_checkDirUserPath . DIRECTORY_SEPARATOR . $logFileName;
            $logFileFullPath = $this->_checkDirFullPath . DIRECTORY_SEPARATOR . $this->getPhase() . '_log';
            try {
                //$userFolder->newFile($logFilePath);
                $this->_logFile = fopen($logFileFullPath, 'w+');
            } catch (Exception $e) {
                $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier de log "' . $logFilePath . '".');
                $this->_msgList['error'][] = $errorMsg;
            }

            $this->_logCsvCols(array('date', 'heure', 'niveau', 'message', 'siret', 'matricule', 'nom_prenom', 'zip_city', 'date_debut', 'date_fin', 'email'));
            $this->_logCsvLine('## ' .$this->_l10n->t('Début du processus de vérification.') . ' ##', self::LOG_DEBUG);

            $csvCheckFileName = $dateTimeNow->format('Y-m-d_H-i-s') . '_' . $uploadFileCSVName;

            $this->_logCsvLine('## ' .$this->_l10n->t('Début de vérification des champs du fichier CSV.') . ' ##', self::LOG_DEBUG);
            // pdf read and parse
            $pdfFilePath = $this->_uploadDirFullPath . DIRECTORY_SEPARATOR . $uploadFilePDFName;
            $pdfPages = array();
            if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                $this->_logCsvLine($this->_l10n->t('Lecture du fichier PDF.'), self::LOG_DEBUG);
                $pdfParser = new \Smalot\PdfParser\Parser();
                $pdf = $pdfParser->parseFile($pdfFilePath);
                $pdfPages = $pdf->getPages();
                $pdfPageNb = count($pdfPages);
                if ($pdfPageNb <= 0) {
                    $errorMsg = $this->_l10n->t('Aucune page dans le PDF "' . $uploadFilePDFName . '".');
                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                    $this->_processErrorMsg(-1, $errorMsg);
                }
            }

            // csv read and load all lines
            $csvTodoLineList = array();
            if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                $this->_logCsvLine($this->_l10n->t('Ouverture du fichier CSV.'), self::LOG_DEBUG);
                $csvFilePath = $this->_uploadDirFullPath . DIRECTORY_SEPARATOR . $uploadFileCSVName;
                $csvFile = fopen($csvFilePath, 'r');
                if (!$csvFile) {
                    $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier CSV "' . $uploadFileCSVName . '".');
                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                    $this->_processErrorMsg(-1, $errorMsg);
                } else {
                    $csvLineNum = 0;
                    $csvLineNumBegin = 0;
                    $this->_csvEncodeFrom = 'ISO-8859-1';
                    $csvConvert = false;
                    if (!empty($this->_csvEncodeFrom) && $this->_csvEncodeFrom!=$this->_csvEncodeTo) {
                        $csvConvert = true;
                    }
                    $this->_logCsvLine($this->_l10n->t('Lecture des lignes du fichier CSV.'), self::LOG_DEBUG);
                    while ($csvLine = fgetcsv($csvFile, $this->_csvLength, $this->_csvDelimiter, $this->_csvEnclosure)) {
                        $this->_logCsvLine('-- ' . $this->_l10n->t('Début de ligne CSV : ' .  ($csvLineNum + 1)) . ' --', self::LOG_DEBUG);

                        if ($csvLineNum >= $csvLineNumBegin) {
                            if ($csvConvert === true) {
                                foreach ($csvLine as $csvColNum => $csvValue) {
                                    $csvLine[$csvColNum] = mb_convert_encoding(trim($csvValue), $this->_csvEncodeTo,  $this->_csvEncodeFrom);
                                }
                            }
                            $matriculeOri = $csvLine[1];
                            $startDate    = date('Y-m-d', strtotime($csvLine[13]));
                            $endDate      = date('Y-m-d', strtotime($csvLine[14]));
                            $editionDate  = date('Y-m-d', strtotime($csvLine[17]));

                            // agent user
                            $agentUser = new DematpayslipUser();
                            $agentUser->setId(0);
                            $agentUser->setUserId('');
                            $agentUser->setSiret($csvLine[0]);
                            $agentUser->setMatricule(str_pad($matriculeOri, '8', '0', STR_PAD_LEFT));
                            $agentUser->setLastname($csvLine[3]);
                            $agentUser->setFirstname($csvLine[4]);
                            $agentUser->setEmail($csvLine[5]);
                            $agentUser->setZipCity($csvLine[10]);

                            // check email field and others fields (regex)
                            $this->_logCsvLine($this->_l10n->t('Ligne CSV'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                            if (empty($agentUser->getEmail())) {
                                $warningMsg = $this->_l10n->t('Le fichier CSV "' . $uploadFileCSVName . '" contient un champ email vide [ligne=' . ($csvLineNum + 1) . '].');
                                $this->_logCsvLine($warningMsg, self::LOG_WARNING, $agentUser, $startDate, $endDate);
                                $this->_processWarningMsg($warningMsg);

                                // skip this line in check and indexation process
                                //$dematpayslipProcessSkipLineArr = $this->_dematpayslipProcessSkipMapper->findAllByIdProcessAndIdLine($this->getId(), $csvLineNum);
                                //if (empty($dematpayslipProcessSkipLineArr)) {
                                //    $dematpayslipProcessSkip = new DematpayslipProcessSkip();
                                //    $dematpayslipProcessSkip->setIdProcess($this->getId());
                                //    $dematpayslipProcessSkip->setIdLine($csvLineNum);
                                //    $this->_dematpayslipProcessSkipMapper->insert($dematpayslipProcessSkip);
                                //}
                            } else {
                                $userFindByEmailList = $this->_userManager->getByEmail($agentUser->getEmail());

                                // /!\ remove all users not available (FIX for multiple users with SAML)
                                if (self::USER_REMOVE_UNKNOWN == 1) {
                                    $userFindByEmailList = self::_removeUnknownUser($userFindByEmailList);
                                }

                                if (count($userFindByEmailList) != 1) {
                                    $warningMsg = $this->_l10n->t('Le fichier CSV "' . $uploadFileCSVName . '" contient un email "' . $agentUser->getEmail() . '" non valide ou non lié à un unique utilisateur Nextcloud [ligne=' . ($csvLineNum + 1) . '].');
                                    $this->_logCsvLine($warningMsg, self::LOG_WARNING, $agentUser, $startDate, $endDate);
                                    $this->_processWarningMsg($warningMsg);

                                    $dematpayslipProcessSkipLineArr = $this->_dematpayslipProcessSkipMapper->findAllByIdProcessAndIdLine($this->getId(), $csvLineNum);
                                    if (empty($dematpayslipProcessSkipLineArr)) {
                                        $dematpayslipProcessSkip = new DematpayslipProcessSkip();
                                        $dematpayslipProcessSkip->setIdProcess($this->getId());
                                        $dematpayslipProcessSkip->setIdLine($csvLineNum);
                                        $this->_dematpayslipProcessSkipMapper->insert($dematpayslipProcessSkip);
                                    }
                                }
                            }

                            if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                                $csvTodoLineList[] = array(
                                    'siret'         => $agentUser->getSiret(),
                                    'matricule_ori' => $matriculeOri,
                                    'lastname'      => $agentUser->getLastname(),
                                    'firstname'     => $agentUser->getFirstname(),
                                    'zip_city'      => $agentUser->getZipCity(),
                                    'start_date'    => $startDate,
                                    'end_date'      => $endDate,
                                    'edition_date'  => $editionDate,
                                    'matricule'     => $agentUser->getMatricule(),
                                    'fullname'      => $agentUser->getLastname() . ' ' . $agentUser->getFirstname(),
                                    'email'         => $agentUser->getEmail(),
                                    'line_num'      => $csvLineNum + 1,
                                    'pdf_pages'     => '',
                                );
                            }
                        }

                        if ($this->_processFailOnFirstError===true && $this->getHasError()>0) {
                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin de ligne CSV : ' .  ($csvLineNum + 1)) . ' --', self::LOG_ERROR);
                            break;
                        } else {
                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin de ligne CSV : ' .  ($csvLineNum + 1)) . ' --', self::LOG_DEBUG);
                        }

                        $csvLineNum++;
                    }
                    fclose($csvFile);
                }
            }
            if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                if (count($csvTodoLineList) <= 0) {
                    $errorMsg = $this->_l10n->t('Aucune ligne dans le fichier CSV "' . $uploadFileCSVName . '".');
                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                    $this->_processErrorMsg(-1, $errorMsg);
                }
            }
            $this->_logCsvLine('## ' .$this->_l10n->t('Fin de vérification des champs du fichier CSV.') . ' ##', self::LOG_DEBUG);

            if ($this->getHasError()>0)   $this->_processFailOnFirstError=true;
            if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                $this->_logCsvLine('## ' .$this->_l10n->t('Début de l\'association des pages du PDF aux différents agents.') . ' ##', self::LOG_DEBUG);

                $this->_logCsvLine($this->_l10n->t('Création du fichier CSV des assoications des pages de bulletins de paie aux différents agents.'), self::LOG_DEBUG);
                $csvCheckFilePath = $this->_checkDirFullPath . DIRECTORY_SEPARATOR . $csvCheckFileName;
                $csvCheckFile = fopen($csvCheckFilePath, 'w+');
                if (!$csvCheckFile) {
                    $errorMsg = $this->_l10n->t('Impossible de créer le fichier CSV "' . $csvCheckFileName . '".');
                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                    $this->_processErrorMsg(-1, $errorMsg);
                }

                if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                    $checkLastValues = array(
                        'siret'     => '',
                        'matricule' => '',
                        'fullname'  => '',
                        'zip_city'  => '',
                    );

                    $checkLastLineNum = -1;

                    // for each PDF page
                    $this->_logCsvLine($this->_l10n->t('Lecture des pages du fichier PDF.'), self::LOG_DEBUG);
                    $pdfPageList = array();
                    foreach ($pdfPages as $pdfPageNum => $pdfPage) {
                        $this->_logCsvLine('-- ' . $this->_l10n->t('Page PDF en cours : ' .  ($pdfPageNum + 1) . ' --'), self::LOG_DEBUG);

                        // get text fields in PDF and remove blanks
                        $this->_logCsvLine($this->_l10n->t('Récupération des champs textes du fichier PDF.'), self::LOG_DEBUG);
                        $pdfPageTextList = array();
                        $pdfPageTextArr = $pdfPage->getTextArray();
                        foreach ($pdfPageTextArr as $pdfPageTextKey => $pdfPageText) {
                            // format text fields (remove blanks)
                            $pdfPageTextList[] = trim($pdfPageText);
                        }

                        // for each CSV line remaining to do
                        $this->_logCsvLine($this->_l10n->t('Lecture des lignes fichier CSV.'), self::LOG_DEBUG);
                        $csvTodoLineNum = 0;
                        $csvTodoLine = array();
                        $checkCurrentValues = array();
                        $pdfPageValues = array();
                        $agentFound = false;
                        foreach ($csvTodoLineList as $csvTodoLineNum => $csvTodoLine) {
                            $this->_logCsvLine('-- ' . $this->_l10n->t('Début de ligne CSV : ' .  ($csvTodoLineNum + 1)) . ' --', self::LOG_DEBUG);

                            // get check values
                            $checkCurrentValues = array(
                                'siret'     => $csvTodoLine['siret'],
                                'matricule' => $csvTodoLine['matricule'],
                                'fullname'  => $csvTodoLine['fullname'],
                                'zip_city'  => $csvTodoLine['zip_city'],
                            );

                            $pdfPageValues = array();

                            // for each value to check
                            foreach ($checkCurrentValues as $checkField => $checkCurrentValue) {
                                // check first in forecast positions in PDF
                                if ($checkField == 'siret') {
                                    // siret
                                    if (isset($this->_payslipPDFSearchFieldList[$checkField]) && !empty($this->_payslipPDFSearchFieldList[$checkField])) {
                                        foreach ($this->_payslipPDFSearchFieldList[$checkField] as $payslipPDFSearchFieldIndex) {
                                            $matchList = array();
                                            if (preg_match('#SIRET:\s(\d+)\s\d+$#', $pdfPageTextList[$payslipPDFSearchFieldIndex], $matchList)) {
                                                if ($matchList[1] == $checkCurrentValue) {
                                                    $pdfPageValues[$checkField] = $checkCurrentValue;
                                                    continue;
                                                }
                                            }
                                        }
                                    }
                                } else {
                                    // siret, matricule, fullname, zip_city
                                    if (isset($this->_payslipPDFSearchFieldList[$checkField]) && !empty($this->_payslipPDFSearchFieldList[$checkField])) {
                                        foreach ($this->_payslipPDFSearchFieldList[$checkField] as $payslipPDFSearchFieldIndex) {
                                            if ($pdfPageTextList[$payslipPDFSearchFieldIndex] == $checkCurrentValue) {
                                                $pdfPageValues[$checkField] = $checkCurrentValue;
                                                break;
                                            }
                                        }
                                    }
                                }

                                if (!isset($pdfPageValues[$checkField])) {
                                    // not fount in forecast position in PDF
                                    // for each field text in PDF page
                                    foreach ($pdfPageTextList as $pdfPageTextKey => $pdfPageText) {
                                        // format text fields in PDF (remove blanks)
                                        $pdfPageTextValue = trim($pdfPageText);

                                        if ($checkField == 'siret') {
                                            $matchList = array();
                                            if (preg_match('#SIRET:\s(\d+)\s\d+$#', $pdfPageTextValue, $matchList)) {
                                                if ($matchList[1] == $checkCurrentValue) {
                                                    $pdfPageValues[$checkField] = $checkCurrentValue;
                                                    break;
                                                }
                                            }
                                        } else {
                                            // siret, matricule, fullname, zip_city
                                            if ($pdfPageTextValue == $checkCurrentValue) {
                                                $pdfPageValues[$checkField] = $checkCurrentValue;
                                                break;
                                            }
                                        }
                                    }
                                }
                            }

                            // check if we found a user agent that match CSV line in progress
                            $agentFound = false;
                            if (!empty($pdfPageValues)) {
                                foreach ($checkCurrentValues as $checkField => $checkCurrentValue) {
                                    if (!isset($pdfPageValues[$checkField]) || empty($pdfPageValues[$checkField])) {
                                        $agentFound = false;
                                        break;
                                    } else {
                                        $agentFound = true;
                                    }
                                }
                            }
                            if ($agentFound === true) {
                                break;
                            }

                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin de ligne CSV : ' .  ($csvTodoLineNum + 1)) . ' --', self::LOG_DEBUG);
                        }

                        if ($agentFound !== true) {
                            $errorMsg = $this->_l10n->t('Aucun agent trouve dans la page ' . ($pdfPageNum + 1) . ' du PDF correspondant aux lignes du fichier CSV.');
                            $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                            $this->_processErrorMsg(-1, $errorMsg);
                            break;
                        } else {
                            // if agent user found and match with PDF page
                            $siret = $csvTodoLine['siret'];
                            $matricule = $csvTodoLine['matricule'];
                            $period = $csvTodoLine['start_date'] . '_' . $csvTodoLine['end_date'];
                            $agentKey = $siret . '-' . $matricule;
                            $this->_logCsvLine($this->_l10n->t('Agent trouvé [siret=' . $siret . ', matricule=' . $matricule . ', period=' . $period . '].'), self::LOG_DEBUG);

                            $newAgent = false;
                            foreach ($checkLastValues as $checkLastField => $checLastValue) {
                                if ($pdfPageValues[$checkLastField] != $checLastValue) {
                                    $newAgent = true;
                                    break;
                                }
                            }

                            if ($newAgent === true) {
                                $this->_logCsvLine($this->_l10n->t('C\'est un nouvel agent.'), self::LOG_DEBUG);
                                $this->_logCsvLine($this->_l10n->t('Ajout de la page ' . ($pdfPageNum + 1) .' du PDF des bulletins de paie de l\'agent.'), self::LOG_DEBUG);

                                $pdfPageList[$agentKey] = array('page_num_list' => array($pdfPageNum + 1), 'period' => $period);

                                if ($checkLastLineNum >= 0) {
                                    // keep and put this line CSV in CSV check file for indexation step (add pages numbers of PDF file)
                                    $checkAgentKey = $csvTodoLineList[$checkLastLineNum]['siret'] . '-' . $csvTodoLineList[$checkLastLineNum]['matricule'];
                                    $csvTodoLineList[$checkLastLineNum]['pdf_pages'] = implode('-', $pdfPageList[$checkAgentKey]['page_num_list']);
                                    fputcsv($csvCheckFile, $csvTodoLineList[$checkLastLineNum], $this->_csvDelimiter, $this->_csvEnclosure);
                                    unset($csvTodoLineList[$checkLastLineNum]); // remove last agent user line
                                }

                                // keep line number linked to last agent
                                $checkLastLineNum = $csvTodoLineNum;

                                // put new values in old values to check
                                $checkLastValues = $checkCurrentValues;
                            } else {
                                $this->_logCsvLine($this->_l10n->t('Ajout de la page ' . ($pdfPageNum + 1) .' du PDF des bulletins de paie de l\'agent.'), self::LOG_DEBUG);
                                $pdfPageList[$agentKey]['page_num_list'][] = $pdfPageNum + 1;
                            }

                            // case of last PDF page
                            if ($pdfPageNum == $pdfPageNb - 1) {
                                if ($checkLastLineNum >= 0) {
                                    // keep and put this line CSV in CSV check file for indexation step (add pages numbers of PDF file)
                                    $checkAgentKey = $csvTodoLineList[$checkLastLineNum]['siret'] . '-' . $csvTodoLineList[$checkLastLineNum]['matricule'];
                                    $csvTodoLineList[$checkLastLineNum]['pdf_pages'] = implode('-', $pdfPageList[$checkAgentKey]['page_num_list']);
                                    fputcsv($csvCheckFile, $csvTodoLineList[$checkLastLineNum], $this->_csvDelimiter, $this->_csvEnclosure);
                                    unset($csvTodoLineList[$checkLastLineNum]); // remove last agent user line
                                }
                            }
                        }

                        $this->_logCsvLine('-- ' . $this->_l10n->t('Fin de page PDF : ' .  ($pdfPageNum + 1) . ' --'), self::LOG_DEBUG);
                    }
                    fclose($csvCheckFile);

                    if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                        // if not all CSV lines was done
                        if (count($csvTodoLineList) > 0) {
                            $errorMsgLines = array();
                            $errorMsgLines[] = $this->_l10n->t('Des lignes du fichier CSV "' . $uploadFileCSVName . '" n\'ont pas été traitées') . ' : ';
                            foreach ($csvTodoLineList as $csvTodoLineNum => $csvTodoLine) {
                                $errorMsgLines[] = ' - ' . $this->_l10n->t('la ligne ' . $csvTodoLine['line_num']);
                            }
                            $this->_logCsvLine(implode("\n", $errorMsgLines), self::LOG_ERROR);
                            $this->_processErrorMsg(-1, implode('<br />', $errorMsgLines));
                        }
                    }
                    $this->_logCsvLine('## ' .$this->_l10n->t('Fin de l\'association des pages du PDF aux différents agents.') . ' ##', self::LOG_DEBUG);
                }

                // Step - cut PDF pages
                if ($this->getHasError()>0)   $this->_processFailOnFirstError = true;
                if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                    $this->_logCsvLine('## ' . $this->_l10n->t('Début de découpage des pages du fichier PDF.') . ' ##', self::LOG_DEBUG);

                    // open CSV file of check directory
                    $this->_logCsvLine($this->_l10n->t('Ouverture des lignes du fichier CSV contenant les pages des bulletins de paie à découper.'), self::LOG_DEBUG);
                    $csvCheckFilePath = $this->_checkDirFullPath . DIRECTORY_SEPARATOR . $csvCheckFileName;
                    $csvCheckFile = fopen($csvCheckFilePath, 'r');
                    if (!$csvCheckFile) {
                        $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier CSV "' . $csvCheckFileName . '".');
                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                        $this->_processErrorMsg(-1, $errorMsg);
                    }

                    if ($this->_processFailOnFirstError===false || $this->getHasError()<=0) {
                        // for each CSV line
                        $this->_logCsvLine($this->_l10n->t('Lecture des lignes fichier CSV.'), self::LOG_DEBUG);
                        $csvCheckLineNum = 0;
                        while ($csvCheckLine = fgetcsv($csvCheckFile, $this->_csvLength, $this->_csvDelimiter, $this->_csvEnclosure)) {
                            $this->_logCsvLine('-- ' . $this->_l10n->t('Début de ligne CSV : ' .  ($csvCheckLineNum + 1)) . ' --', self::LOG_DEBUG);
                            //$csvCheckArr[] = array(
                            //    'siret'         => $csvCheckLine[0],
                            //    'matricule_ori' => $csvCheckLine[1],
                            //    'lastname'      => $csvCheckLine[2],
                            //    'firstname'     => $csvCheckLine[3],
                            //    'zip_city'      => $csvCheckLine[4],
                            //    'start_date'    => $csvCheckLine[5],
                            //    'end_date'      => $csvCheckLine[6],
                            //    'edition_date'  => $csvCheckLine[7],
                            //    'matricule'     => $csvCheckLine[8],
                            //    'fullname'      => $csvCheckLine[9],
                            //    'email'         => $csvCheckLine[10],
                            //    'line_num'      => $csvCheckLine[11],
                            //    'pdf_pages'     => $csvCheckLine[12],
                            //);

                            // get CSV data values
                            $csvCheckMatricule = $csvCheckLine[8];
                            $pdfPageNumList = explode('-', $csvCheckLine[12]);
                            $this->_logCsvLine($this->_l10n->t('Découpage des pages du PDF pour le bulletin de paie de l\'agent [matricule=' . $csvCheckMatricule . '].'), self::LOG_DEBUG);

                            // add pages
                            $this->_logCsvLine($this->_l10n->t('Ajout des pages PDF au bulletin de paie.'), self::LOG_DEBUG);
                            $pdfPayslip = new \setasign\Fpdi\Fpdi();
                            foreach ($pdfPageNumList as $pdfPageNum) {
                                $this->_logCsvLine($this->_l10n->t('Ajout de la page ' . $pdfPageNum . '.'), self::LOG_DEBUG);
                                $pdfPayslip->AddPage();
                                $pdfPayslip->setSourceFile($pdfFilePath);
                                $tplPayslipId = $pdfPayslip->importPage($pdfPageNum);
                                $pdfPayslip->useTemplate($tplPayslipId);
                            }

                            // put payslip in check directory
                            $this->_logCsvLine($this->_l10n->t('Créer le PDF du bulletin de paie.'), self::LOG_DEBUG);
                            $pdfPayslip->Output('F', $this->_checkDirFullPath . DIRECTORY_SEPARATOR . ($csvCheckLineNum + 1) . '_' . $csvCheckMatricule . '.pdf');

                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin de ligne CSV : ' .  ($csvCheckLineNum + 1)) . ' --', self::LOG_DEBUG);

                            $csvCheckLineNum++;
                        }
                        fclose($csvCheckFile);
                    }

                    $this->_logCsvLine('## ' . $this->_l10n->t('Fin de découpage des pages du fichier PDF.') . ' ##', self::LOG_DEBUG);
                }
            }

            if ($this->getHasError() <= 0) {
                $successMsg = $this->_l10n->t('Opération "' . self::_getPhaseName($this->getPhase()) . '" terminée avec succès.');
                $this->_logCsvLine($successMsg, self::LOG_SUCCESS);
                $this->_msgList['success'][] = $successMsg;
                $this->setPhase(self::PHASE_ID_PRE_INDEXATION);
                $this->setCsvCheck($csvCheckFileName);
            }

            $this->_logCsvLine('## ' .$this->_l10n->t('Fin du processus de vérification.') . ' ##', self::LOG_DEBUG);
            fclose($this->_logFile);
            // rename (log file)
            try {
                $logFileNode = $userFolder->get($logFilePath);
                $logFileNode->move($logFileNode->getPath() . '.csv');
            } catch (Exception $e) {
                $this->_msgList['error'][] = $this->_l10n->t('Impossible de déplacer le fichier de log "' . $logFilePath . '" .');
            }
            $this->setProcessing(0);
            $this->_dematpayslipProcessMapper->update($this);
        }
    }

    /**
     * Pre-indexation
     */
    public function preIndexation() {
        $this->_initDematUser();

        $this->_processCheck();
        if ($this->getPhase() != self::PHASE_ID_PRE_INDEXATION) {
            $this->_msgList['error'][] = $this->_l10n->t('Opération impossible : phase en cours "' . self::_getPhaseName($this->getPhase()) . '".');
        }

        $launchIndexation = filter_input(INPUT_POST, 'payslip_launch_indexation', FILTER_SANITIZE_NUMBER_INT);
        $launchArchive = filter_input(INPUT_POST, 'payslip_launch_archive', FILTER_SANITIZE_NUMBER_INT);
        $launchIndexation = intval($launchIndexation);
        $launchArchive = intval($launchArchive);
        if ($launchIndexation != 1) {
            $this->_msgList['error'][] = $this->_l10n->t('Veuillez sélectionner au moins l\'étape d\'"Indexation".');
        }

        if (!$this->hasError()) {
            $operationAskList = array();
            if ($launchIndexation == 1) {
                $operationAskList[] = DematpayslipArchive::OPERATION_INDEXATION;
            }
            if ($launchArchive == 1) {
                $operationAskList[] = DematpayslipArchive::OPERATION_ARCHIVE;
            }
            $operationAsk = implode('+', $operationAskList);

            try {
                $phaseStartDateTime = new DateTime('now', $this->_timeZone);
                $phaseStartTime = $phaseStartDateTime->getTimestamp();

                $this->setProcessing(1);
                $this->setPhase(self::PHASE_ID_PRE_INDEXATION);
                $this->setPhaseStartTime($phaseStartTime);
                $this->setPhaseStepNum(0);
                $this->setMustTerminate(0);
                $this->setHasError(0);
                $this->setErrorCode(0);
                $this->setErrorMsg('');
                $this->setUserId($this->_userId);
                $this->_dematpayslipProcessMapper->update($this);

                $payslipsDirPath = (string) $this->_dematpayslipConfigList['payslip_dir'];
                $csvCheckFileName = $this->getCsvCheck();

                // check and mount user directory
                $userHomeFilesDirPath = '';
                if ($this->getHasError() <= 0) {
                    $user = $this->_userManager->get($this->_userId);
                    if (!$user) {
                        $errorMsg = $this->_l10n->t('Le dossier utilisateur "' . $this->_userId . '" n\'existe pas.');
                        $this->_processErrorMsg(-1, $errorMsg);
                    }

                    $userHomeFilesDirPath = $user->getHome() . DIRECTORY_SEPARATOR . 'files';
                }
                if (empty($userHomeFilesDirPath) || !is_dir($userHomeFilesDirPath)) {
                    $errorMsg = $this->_l10n->t('Le dossier utilisateur "' . $this->_userId . '" est introuvable.');
                    $this->_processErrorMsg(-1, $errorMsg);
                }
                $userFolder = $this->_rootFolder->getUserFolder($this->_userId);

                // check user payslip dir exists
                if ($this->getHasError() <= 0) {
                    try {
                        $userFolder->get($payslipsDirPath);
                    } catch (Exception $e) {
                        $errorMsg = $this->_l10n->t('Le dossier des bulletins "' . $payslipsDirPath . '" n\'existe pas.');
                        $this->_processErrorMsg(-1, $errorMsg);
                    }
                }

                // check "no indexed payslip" directory exists
                $this->_logCsvLine($this->_l10n->t('Vérification de l\'existance du dossier des bulletins non indexes.'), self::LOG_DEBUG);
                if ($this->getHasError() <= 0) {
                    try {
                        $userFolder->get($this->_payslipNoIndexDir);
                    } catch (Exception $e) {
                        $errorMsg = $this->_l10n->t('Le dossier des bulletins non indexes "' . $this->_payslipNoIndexDir . '" n\'existe pas.');
                        $this->_processErrorMsg(-1, $errorMsg);
                    }
                }

                if ($this->getHasError() <= 0) {
                    $phaseStep = 'csv_check';
                    $csvCheckLineNum = 0;
                    if (empty($this->getPhaseStep())) {
                        // get all share with this user
                        /** @var Folder $payslipShareNode */
                        $payslipShareNode = $userFolder->get($payslipsDirPath);
                        $payslipShareList = $this->_shareManager->getSharesInFolder($this->_userId, $payslipShareNode);
                        $payslipMatriculeShareNodeIdList = array();
                        if (count($payslipShareList) > 0) {
                            foreach ($payslipShareList as $payslipShareNodeId => $payslipShareArr) {
                                $payslipMatriculeShareNodeIdList[] = $payslipShareNodeId;
                            }
                        }

                        // open CSV file created in check step
                        $csvCheckFilePath = $this->_checkDirFullPath . DIRECTORY_SEPARATOR . $csvCheckFileName;
                        $csvCheckFile = fopen($csvCheckFilePath, 'r');
                        if (!$csvCheckFile) {
                            $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier CSV "' . $csvCheckFileName . '".');
                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                        }

                        if ($this->getHasError() <= 0) {
                            // get CSV lines to skip
                            $dematpayslipProcessSkipLineIdList = array();
                            $dematpayslipProcessSkipList = $this->_dematpayslipProcessSkipMapper->findAllByIdProcess($this->getId());
                            foreach ($dematpayslipProcessSkipList as $dematpayslipProcessSkip) {
                                $dematpayslipProcessSkipLineIdList[] = $dematpayslipProcessSkip->getIdLine();
                            }

                            // payslip archive mapper
                            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

                            while ($csvCheckLine = fgetcsv($csvCheckFile, $this->_csvLength, $this->_csvDelimiter, $this->_csvEnclosure)) {
                                //$csvCheckArr[] = array(
                                //    'siret'         => $csvCheckLine[0],
                                //    'matricule_ori' => $csvCheckLine[1],
                                //    'lastname'      => $csvCheckLine[2],
                                //    'firstname'     => $csvCheckLine[3],
                                //    'zip_city'      => $csvCheckLine[4],
                                //    'start_date'    => $csvCheckLine[5],
                                //    'end_date'      => $csvCheckLine[6],
                                //    'edition_date'  => $csvCheckLine[7],
                                //    'matricule'     => $csvCheckLine[8],
                                //    'fullname'      => $csvCheckLine[9],
                                //    'email'         => $csvCheckLine[10],
                                //    'line_num'      => $csvCheckLine[11],
                                //    'pdf_pages'     => $csvCheckLine[12],
                                //);

                                // get CSV data values
                                $startDate = $csvCheckLine[5];
                                $endDate = $csvCheckLine[6];
                                $editionDate = $csvCheckLine[7];
                                $startDateTime = strtotime($startDate);
                                $year = date('Y', $startDateTime);
                                $month = strtolower(strftime('%B', $startDateTime));
                                $pdfPagesNum = $csvCheckLine[12];

                                // agent user
                                $agentUser = new DematpayslipUser();
                                $agentUser->setId(0);
                                $agentUser->setUserId('');
                                $agentUser->setSiret($csvCheckLine[0]);
                                $agentUser->setMatricule($csvCheckLine[8]);
                                $agentUser->setLastname($csvCheckLine[2]);
                                $agentUser->setFirstname($csvCheckLine[3]);
                                $agentUser->setEmail($csvCheckLine[10]);
                                $agentUser->setZipCity($csvCheckLine[4]);

                                // pdf check file path
                                $pdfPayslipCheckFileName = ($csvCheckLineNum + 1) . '_' . $agentUser->getMatricule() . '.pdf';
                                $pdfPayslipCheckFilePath = $this->_checkDirUserPath . DIRECTORY_SEPARATOR . $pdfPayslipCheckFileName;
                                /** @var File $pdfPayslipCheckFileNode */
                                $pdfPayslipCheckFileNode = $userFolder->get($pdfPayslipCheckFilePath);

                                // process skip lines
                                if (in_array($csvCheckLineNum, $dematpayslipProcessSkipLineIdList)) {
                                    // check if not already archived (return null if already archived with warning)
                                    $checkArchive = $this->_archiveCheckBeforeSave($dematpayslipArchiveMapper, $agentUser, $startDate, $endDate);
                                    if (!empty($checkArchive)) {
                                        // save PDF temp file path and activate indexation job
                                        $checkArchive->saveFromFilePathAndAgent($pdfPayslipCheckFileNode, $dematpayslipArchiveMapper, $agentUser, $startDate, $endDate, $year, $month, $editionDate, DematpayslipArchive::STATUS_INITIALIZED, null, $this->getId(), $pdfPagesNum, $operationAsk, DematpayslipArchive::OPERATION_INDEXATION);
                                    }

                                    $csvCheckLineNum++;
                                    continue;
                                }

                                // no user but archiving for 5 years "BulletinsNoIndex/HORODATE/SIRET-MATRICULE/PERIODE.pdf"
                                if (empty($agentUser->getEmail())) {
                                    // check if not already archived (return null if already archived with warning)
                                    $checkArchive = $this->_archiveCheckBeforeSave($dematpayslipArchiveMapper, $agentUser, $startDate, $endDate);
                                    if (!empty($checkArchive)) {
                                        // save PDF temp file path and activate indexation job
                                        $checkArchive->saveFromFilePathAndAgent($pdfPayslipCheckFileNode, $dematpayslipArchiveMapper, $agentUser, $startDate, $endDate, $year, $month, $editionDate, DematpayslipArchive::STATUS_INITIALIZED, null, $this->getId(), $pdfPagesNum, $operationAsk, DematpayslipArchive::OPERATION_INDEXATION);
                                    }

                                    $csvCheckLineNum++;
                                    continue;
                                }

                                // get agent from email
                                $userFindByEmailList = $this->_userManager->getByEmail($agentUser->getEmail());

                                // /!\ remove all users not available (FIX for multiple users with SAML)
                                if (self::USER_REMOVE_UNKNOWN == 1) {
                                    $userFindByEmailList = self::_removeUnknownUser($userFindByEmailList);
                                }

                                if (count($userFindByEmailList) != 1) {
                                    $errorMsg = $this->_l10n->t('Impossible de trouver l\'unique utilisateur Nextcloud dont l\'email est "' . $agentUser->getEmail() . '" [matricule="' . $agentUser->getMatricule() . '"].');
                                    $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                }
                                $userFindByEmail = current($userFindByEmailList);
                                $userFindByEmailId = $userFindByEmail->getUID();

                                if ($this->getHasError() <= 0) {
                                    // "SIRET-MATRICULE" directory
                                    $siretMatriculeDirName = $agentUser->getSiret() . '-' . $agentUser->getMatricule();
                                    $payslipsMatriculeDirPath = $payslipsDirPath . DIRECTORY_SEPARATOR . $siretMatriculeDirName;

                                    try {
                                        $userFolder->get($payslipsMatriculeDirPath);
                                    } catch (Exception $e) {
                                        $createdFolder = $userFolder->newFolder($payslipsMatriculeDirPath);
                                        if (!$createdFolder) {
                                            $errorMsg = $this->_l10n->t('Impossible de créer le dossier "' . $payslipsMatriculeDirPath . '" [matricule="' . $agentUser->getMatricule() . '"].');
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                        }

                                        // create user linked to this payslip
                                        $dematpayslipUserList = $this->_dematpayslipUserMapper->findAllBySiretAndMatricule($agentUser->getSiret(), $agentUser->getMatricule());
                                        if (empty($dematpayslipUserList)) {
                                            $dematpayslipUser = new DematpayslipUser();
                                            $dematpayslipUser->setUserId($userFindByEmailId);
                                            $dematpayslipUser->setSiret($agentUser->getSiret());
                                            $dematpayslipUser->setMatricule($agentUser->getMatricule());
                                            $dematpayslipUser->setEmail($agentUser->getEmail());
                                            $dematpayslipUser->setLastname($agentUser->getLastname());
                                            $dematpayslipUser->setFirstname($agentUser->getFirstname());
                                            $dematpayslipUser->setZipCity($agentUser->getZipCity());
                                            $this->_dematpayslipUserMapper->insert($dematpayslipUser);
                                        } else {
                                            $errorMsg = $this->_l10n->t('L\'utilisateur "' . $userFindByEmailId . '" est déjà associé à un dossier SIRET-MATRICULE [siret="' . $agentUser->getSiret() . '", matricule="' . $agentUser->getMatricule() . '", email="' . $agentUser->getEmail() . '"].');
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                        }
                                    }

                                    if ($this->getHasError() <= 0) {
                                        // get created user and linked to "SIRET-MATRICULE" directory
                                        $dematpayslipUserList = $this->_dematpayslipUserMapper->findAllBySiretAndMatricule($agentUser->getSiret(), $agentUser->getMatricule());
                                        if (count($dematpayslipUserList) === 1) {
                                            $dematpayslipUser = $dematpayslipUserList[0];

                                            // set agent user id
                                            $agentUser->setId($dematpayslipUser->getId());
                                            $agentUser->setUserId($dematpayslipUser->getUserId());

                                            // check if not already archived (return null if already archived with warning)
                                            $checkArchive = $this->_archiveCheckBeforeSave($dematpayslipArchiveMapper, $agentUser, $startDate, $endDate);
                                            if (!empty($checkArchive)) {
                                                // save PDF temp file path and activate indexation job
                                                $checkArchive->saveFromFilePathAndAgent($pdfPayslipCheckFileNode, $dematpayslipArchiveMapper, $agentUser, $startDate, $endDate, $year, $month, $editionDate, DematpayslipArchive::STATUS_INITIALIZED, null, $this->getId(), $pdfPagesNum, $operationAsk, DematpayslipArchive::OPERATION_INDEXATION);
                                            }
                                        } else {
                                            $errorMsg = $this->_l10n->t('L\'utilisateur "' . $userFindByEmailId . '" n\'a pas de dossier SIRET-MATRICULE associé [siret="' . $agentUser->getSiret() . '", matricule="' . $agentUser->getMatricule() . '", email="' . $agentUser->getEmail() . '"].');
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                        }
                                    }
                                }

                                if ($this->getHasError() > 0) {
                                    break;
                                }

                                $csvCheckLineNum++;
                            }
                            fclose($csvCheckFile);
                        }
                    }
                }

                if ($this->getHasError() <= 0) {
                    $successMsg = 'Opération "' . self::_getPhaseName($this->getPhase()) . '" terminée avec succès.';
                    $this->_msgList['success'][] = $successMsg;
                }
                $this->setProcessing(0);
                $this->setPhase(self::PHASE_ID_INDEXATION);
                $this->_dematpayslipProcessMapper->update($this);
            } catch (Exception $e) {
                $errorMsg = 'Exception ' . __METHOD__ . ' : error=' . $e->getMessage();
                $this->_msgList['error'][] = $errorMsg;
            }
        }
    }

    /**
     * Indexation
     *
     * @param   array   $processMsgList     Process message list
     * @return  int                         <0 if errors else >0 on success
     */
    public function indexation(&$processMsgList) {
        try {
            $this->_initDematUser();

            $retry = 0;
            if ($this->getHasError() > 0) {
                $retry = $this->getPhaseStepNum();
            }
            if ($retry > 0) {
                $this->setHasError(0);
                $this->setErrorCode(0);
                $this->setErrorMsg('');
            }

            $this->_processCheck();
            if ($this->getPhase() != self::PHASE_ID_INDEXATION) {
                $this->_msgList['error'][] = $this->_l10n->t('Opération impossible : phase en cours "' . self::_getPhaseName($this->getPhase()) . '".');
            }

            if (!$this->hasError()) {
                $phaseStartDateTime = new DateTime('now', $this->_timeZone);
                $phaseStartTime = $phaseStartDateTime->getTimestamp();

                // log file
                $logFileName = $phaseStartDateTime->format('Y-m-d_H-i-s') . '_' . $this->getPhase() . '_log';
                $logFilePath = $this->_reportDirUserPath . DIRECTORY_SEPARATOR . $logFileName;
                $logFileFullPath = $this->_reportDirFullPath . DIRECTORY_SEPARATOR . $logFileName;
                $this->_logFile = fopen($logFileFullPath, 'w+');
                $this->_logCsvCols(array('date', 'heure', 'niveau', 'message', 'siret', 'matricule', 'nom_prenom', 'zip_city', 'date_debut', 'date_fin', 'email'));
                $this->_logCsvLine('## ' . $this->_l10n->t('Début du processus d\'indexation.') . ' ##', self::LOG_DEBUG);

                $this->setProcessing(1);
                $this->setPhase(self::PHASE_ID_INDEXATION);
                $this->setPhaseStartTime($phaseStartTime);
                if (!($retry > 0)) {
                    $this->setPhaseStepNum(0);
                }
                $this->setMustTerminate(1);
                $this->setHasError(0);
                $this->setErrorCode(0);
                $this->setErrorMsg('');
                $this->setUserId($this->_userId);
                $this->_dematpayslipProcessMapper->update($this);

                $payslipsDirPath = (string) $this->_dematpayslipConfigList['payslip_dir'];
                $csvCheckFileName = $this->getCsvCheck();

                // check and mount user directory
                $this->_logCsvLine($this->_l10n->t('Récupération et montage du dossier de l\'utilisateur Nextcloud.'), self::LOG_DEBUG);
                $userHomeFilesDirPath = '';
                if (!$this->hasError()) {
                    $user = $this->_userManager->get($this->_userId);
                    if (!$user) {
                        $errorMsg = $this->_l10n->t('Le dossier utilisateur "' . $this->_userId . '" n\'existe pas.');
                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                        $this->_processErrorMsg(-1, $errorMsg);
                    }

                    $userHomeFilesDirPath = $user->getHome() . DIRECTORY_SEPARATOR . 'files';
                }
                if (empty($userHomeFilesDirPath) || !is_dir($userHomeFilesDirPath)) {
                    $errorMsg = $this->_l10n->t('Le dossier utilisateur "' . $this->_userId . '" est introuvable.');
                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                    $this->_processErrorMsg(-1, $errorMsg);
                }
                $userFolder = $this->_rootFolder->getUserFolder($this->_userId);

                // check user payslip dir exists
                $this->_logCsvLine($this->_l10n->t('Vérification de l\'existance du dossier des bulletins.'), self::LOG_DEBUG);
                if (!$this->hasError()) {
                    try {
                        $userFolder->get($payslipsDirPath);
                    } catch (Exception $e) {
                        $errorMsg = $this->_l10n->t('Le dossier des bulletins "' . $payslipsDirPath . '" n\'existe pas.');
                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                        $this->_processErrorMsg(-1, $errorMsg);
                    }
                }

                // check "no indexed payslip" directory exists
                $this->_logCsvLine($this->_l10n->t('Vérification de l\'existance du dossier des bulletins non indexes.'), self::LOG_DEBUG);
                if (!$this->hasError()) {
                    try {
                        $userFolder->get($this->_payslipNoIndexDir);
                    } catch (Exception $e) {
                        $errorMsg = $this->_l10n->t('Le dossier des bulletins non indexes "' . $this->_payslipNoIndexDir . '" n\'existe pas.');
                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                        $this->_processErrorMsg(-1, $errorMsg);
                    }
                }

                if (!$this->hasError()) {
                    $phaseStep = 'indexation';
                    $this->_logCsvLine($this->_l10n->t('Début de la phase "' . $phaseStep . '".'), self::LOG_DEBUG);
                    $csvCheckLineNum = 0;
                    $retryCsvNumLine = 0;
                    if ($retry > 0) {
                        if ($this->getPhaseStep() == $phaseStep) {
                            $this->setPhaseStep('');
                        }
                        $retryCsvNumLine = $this->getPhaseStepNum();
                        $this->_logCsvLine($this->_l10n->t('Reprise à la ligne ' . ($retryCsvNumLine + 1) . ' du fichier CSV créé lors de la phase de vérification.'), self::LOG_DEBUG);
                    }

                    if (empty($this->getPhaseStep())) {
                        $userFolder = $this->_rootFolder->getUserFolder($this->_userId);

                        // get all share with this user
                        $this->_logCsvLine($this->_l10n->t('Récupération des partages de "' . $payslipsDirPath . '".'), self::LOG_DEBUG);
                        /** @var Folder $payslipShareNode */
                        $payslipShareNode = $userFolder->get($payslipsDirPath);
                        $payslipShareList = $this->_shareManager->getSharesInFolder($this->_userId, $payslipShareNode);
                        $payslipMatriculeShareNodeIdList = array();
                        if (count($payslipShareList) > 0) {
                            foreach ($payslipShareList as $payslipShareNodeId => $payslipShareArr) {
                                $payslipMatriculeShareNodeIdList[] = $payslipShareNodeId;
                            }
                        }

                        if (!$this->hasError()) {
                            $this->_logCsvLine($this->_l10n->t('Lecture du fichier PDF.'), self::LOG_DEBUG);
                            $pdfOriginFileFullPath = $this->_uploadDirFullPath . DIRECTORY_SEPARATOR . $this->getPdfOrigin();

                            // create PDF payslip to skip (not indexed)
                            $pdfPayslipSkip = null;
                            $payslipSkipPDFPath = $this->_payslipNoIndexDir . DIRECTORY_SEPARATOR . $phaseStartDateTime->format('Y-m-d_H-i-s') . '_skip';
                            $skipArchiveList = array();

                            // get CSV lines to skip
                            $dematpayslipProcessSkipLineIdList = array();
                            $dematpayslipProcessSkipList = $this->_dematpayslipProcessSkipMapper->findAllByIdProcess($this->getId());
                            foreach ($dematpayslipProcessSkipList as $dematpayslipProcessSkip) {
                                $dematpayslipProcessSkipLineIdList[] = $dematpayslipProcessSkip->getIdLine();
                            }

                            // payslip archive mapper
                            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

                            // find all payslip files to be indexed
                            $this->_logCsvLine($this->_l10n->t('Lecture de la liste des bulletins à traiter.'), self::LOG_DEBUG);
                            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAllByDematIdProcessAndStatus($this->getId(), DematpayslipArchive::STATUS_INITIALIZED);
                            if (!empty($dematpayslipArchiveList)) {
                                foreach ($dematpayslipArchiveList as $dematpayslipArchive) {
                                    $archiveError = 0;

                                    // only for operation ask
                                    $operationAsk = $dematpayslipArchive->getOperationAsk();
                                    $operationAskList = explode('+', $operationAsk);
                                    if (in_array(DematpayslipArchive::OPERATION_INDEXATION, $operationAskList)) {
                                        $this->_logCsvLine('-- ' . $this->_l10n->t('Bulletin en cours : ' . $dematpayslipArchive->getFilePath() . ' [Id = ' . $dematpayslipArchive->getId(). ', File id = ' . $dematpayslipArchive->getIdFile() . ']') . ' --', self::LOG_DEBUG);

                                        $operationAskArchive = in_array(DematpayslipArchive::OPERATION_ARCHIVE, $operationAskList);

                                        // get payslip file values
                                        $startDate = $dematpayslipArchive->getStartDate();
                                        $endDate =  $dematpayslipArchive->getEndDate();
                                        $editionDate = $dematpayslipArchive->getEditionDate();
                                        $year = $dematpayslipArchive->getYear();;
                                        $month = $dematpayslipArchive->getMonth();;
                                        $pdfPageNumList = explode('-', $dematpayslipArchive->getFilePagesNum());
                                        $period = $startDate . '_' . $endDate;

                                        // agent user
                                        $agentUser = null;
                                        if ($dematpayslipArchive->getDematpayslipIdUser() > 0) {
                                            $agentUser = $this->_dematpayslipUserMapper->find($dematpayslipArchive->getDematpayslipIdUser());
                                        }
                                        if (!$agentUser) {
                                            $agentUser = new DematpayslipUser();
                                            $agentUser->setId(0);
                                            $agentUser->setUserId('');
                                            $agentUser->setSiret($dematpayslipArchive->getSiret());
                                            $agentUser->setMatricule($dematpayslipArchive->getMatricule());
                                            $agentUser->setLastname($dematpayslipArchive->getLastname());
                                            $agentUser->setFirstname($dematpayslipArchive->getFirstname());
                                            $agentUser->setEmail($dematpayslipArchive->getEmail());
                                            $agentUser->setZipCity($dematpayslipArchive->getZipCity());
                                        }

                                        // pdf payslip file path
                                        $pdfPayslipCheckFilePath = substr($dematpayslipArchive->getFilePath(), strlen($userFolder->getPath() . DIRECTORY_SEPARATOR)); // remove "/user/files/" in archive file path
                                        $pdfPayslipCheckFileName = basename($pdfPayslipCheckFilePath);
                                        $pdfCheckFileNames = explode('_', $pdfPayslipCheckFileName);
                                        $csvCheckLineNum = intval($pdfCheckFileNames[0]) - 1;
                                        if ($csvCheckLineNum < $retryCsvNumLine)    continue; // already done

                                        $this->_logCsvLine($this->_l10n->t('Bulletin de l\'agent '), self::LOG_DEBUG, $agentUser, $startDate, $endDate);

                                        // process skip lines
                                        if (in_array($csvCheckLineNum, $dematpayslipProcessSkipLineIdList)) {
                                            $this->_logCsvLine('-- ' . $this->_l10n->t('Bulletin à ne pas indexer : ' . ($csvCheckLineNum + 1)) . ' --', self::LOG_DEBUG, $agentUser, $startDate, $endDate);

                                            // create PDF for payslip not to be indexed
                                            if ($pdfPayslipSkip === null) {
                                                $pdfPayslipSkip = new \setasign\Fpdi\Fpdi();
                                            }
                                            // add pages in user payslip not to be indexed
                                            $this->_logCsvLine($this->_l10n->t('Ajout des pages PDF au bulletin de paie à ne pas indexer.'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                            foreach ($pdfPageNumList as $pdfPageNum) {
                                                $this->_logCsvLine($this->_l10n->t('Ajout de la page ' . $pdfPageNum . '.'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                $pdfPayslipSkip->AddPage();
                                                $pdfPayslipSkip->setSourceFile($pdfOriginFileFullPath);
                                                $tplPayslipSkipId = $pdfPayslipSkip->importPage($pdfPageNum);
                                                $pdfPayslipSkip->useTemplate($tplPayslipSkipId);
                                            }

                                            // not archiving (skip)
                                            $skipArchiveList[] = $dematpayslipArchive;

                                            continue;
                                        }

                                        // no user but archiving for 5 years "BulletinsNoIndex/HORODATE/SIRET-MATRICULE/PERIODE.pdf"
                                        if (empty($agentUser->getEmail())) {
                                            $this->_logCsvLine('-- ' . $this->_l10n->t('Création du PDF non dématérialisé de l\'agent (aucun email) : ' . ($csvCheckLineNum + 1)) . ' --', self::LOG_DEBUG, $agentUser, $startDate, $endDate);

                                            // create PDF payslip
                                            $payslipNoDematHorodateDirPath = $this->_payslipNoIndexDir . DIRECTORY_SEPARATOR . $phaseStartDateTime->format('Y-m-d_H-i-s');
                                            try {
                                                $userFolder->get($payslipNoDematHorodateDirPath);
                                            } catch (Exception $e) {
                                                $createdFolder = $userFolder->newFolder($payslipNoDematHorodateDirPath);
                                                if (!$createdFolder) {
                                                    $archiveError++;
                                                    $errorMsg = $this->_l10n->t('Impossible de créeer le dossier "' . $payslipNoDematHorodateDirPath . '".');
                                                    $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                    $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                }
                                            }
                                            $payslipNoDematHorodateAgentDirPath = $payslipNoDematHorodateDirPath . DIRECTORY_SEPARATOR . $agentUser->getSiret() . '-' . $agentUser->getMatricule();
                                            try {
                                                $userFolder->get($payslipNoDematHorodateAgentDirPath);
                                            } catch (Exception $e) {
                                                $createdFolder = $userFolder->newFolder($payslipNoDematHorodateAgentDirPath);
                                                if (!$createdFolder) {
                                                    $archiveError++;
                                                    $errorMsg = $this->_l10n->t('Impossible de créeer le dossier "' . $payslipNoDematHorodateAgentDirPath . '".');
                                                    $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                    $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                }
                                            }
                                            $payslipNoDematPDFPath = $payslipNoDematHorodateAgentDirPath . DIRECTORY_SEPARATOR . $startDate . '_' . $endDate;

                                            // create PDF payslip in no index folder
                                            $this->_logCsvLine($this->_l10n->t('Créer le PDF du bulletin de paie non dematérialisé.'), self::LOG_DEBUG);
                                            $pdfPayslipFilePath = $payslipNoDematPDFPath . '.pdf';
                                            try {
                                                /** @var File $payslipNoDematPDFNode */
                                                $payslipNoDematPDFNode = $userFolder->get($pdfPayslipCheckFilePath);
                                                $payslipNoDematPDFNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath);
                                                $dematpayslipArchive->setIdFile($payslipNoDematPDFNode->getId());
                                                $dematpayslipArchive->setFilePath($payslipNoDematPDFNode->getPath());
                                                $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                            } catch (Exception $e) {
                                                $archiveError++;
                                                $errorMsg = $this->_l10n->t('Impossible de déplacer le PDF du bulletin de paie non dematérialisé "' . $userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath . '".');
                                                $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                                $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                            }

                                            if ($archiveError > 0) {
                                                // file finished with errors
                                                $processMsgList['error'][] = $dematpayslipArchive;
                                                $dematpayslipArchive->setError(1);
                                                $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                            } else {
                                                $processMsgList['success'][] = $dematpayslipArchive;
                                                $dematpayslipArchive->setStatus(DematpayslipArchive::STATUS_INDEXED);
                                                $dematpayslipArchive->setStatusTime($phaseStartTime);
                                                // next operation for this file if asked
                                                if ($operationAskArchive) {
                                                    $dematpayslipArchive->setOperationProgress(DematpayslipArchive::OPERATION_ARCHIVE);
                                                } else {
                                                    $dematpayslipArchive->setOperationProgress(DematpayslipArchive::OPERATION_FINISHED);
                                                }
                                                $dematpayslipArchive->setError(0);
                                                $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                            }

                                            continue;
                                        }

                                        // get agent from email
                                        $this->_logCsvLine($this->_l10n->t('Récupération de l\'utilisateur Nextcloud à partir de son email [' . $agentUser->getEmail() . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                        $userFindByEmailList = $this->_userManager->getByEmail($agentUser->getEmail());

                                        // /!\ remove all users not available (FIX for multiple users with SAML)
                                        if (self::USER_REMOVE_UNKNOWN == 1) {
                                            $userFindByEmailList = self::_removeUnknownUser($userFindByEmailList);
                                        }

                                        if (count($userFindByEmailList) != 1) {
                                            $archiveError++;
                                            $errorMsg = $this->_l10n->t('Impossible de trouver l\'unique utilisateur Nextcloud dont l\'email est "' . $agentUser->getEmail() . '" [matricule="' . $agentUser->getMatricule() . '"].');
                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                        }
                                        $userFindByEmail = current($userFindByEmailList);
                                        $userFindByEmailId = $userFindByEmail->getUID();

                                        if (!$this->hasError()) {
                                            // "SIRET-MATRICULE" directory
                                            $siretMatriculeDirName = $agentUser->getSiret() . '-' . $agentUser->getMatricule();
                                            $payslipsMatriculeDirPath = $payslipsDirPath . DIRECTORY_SEPARATOR . $siretMatriculeDirName;

                                            // get agent user from "SIRET-MATRICULE" created directory
                                            $this->_logCsvLine($this->_l10n->t('Récupération de l\'utilisateur associé au dossier SIRET-MATRICULE [' . $siretMatriculeDirName . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                            $dematpayslipUserList = $this->_dematpayslipUserMapper->findAllBySiretAndMatricule($agentUser->getSiret(), $agentUser->getMatricule());
                                            if (count($dematpayslipUserList) === 1) {
                                                $dematpayslipUser = current($dematpayslipUserList);

                                                // update user for new version
                                                if (empty($dematpayslipUser->getLastname()) || empty($dematpayslipUser->getFirstname()) || empty($dematpayslipUser->getZipCity())) {
                                                    $dematpayslipUser->setLastname($agentUser->getLastname());
                                                    $dematpayslipUser->setFirstname($agentUser->getFirstname());
                                                    $dematpayslipUser->setZipCity($agentUser->getZipCity());
                                                    $this->_dematpayslipUserMapper->update($dematpayslipUser);
                                                }

                                                // set agent user id
                                                $agentUser->setId($dematpayslipUser->getId());
                                                $agentUser->setUserId($dematpayslipUser->getUserId());

                                                // chek if user disabled payslip
                                                if ($dematpayslipUser->getDisabled() != 1) {
                                                    $payslipsMatriculeYearDirPath = $payslipsMatriculeDirPath . DIRECTORY_SEPARATOR . $year;

                                                    // create share directory "SIRET-MATRICULE"
                                                    $payslipsMatriculeDirPathNode = $userFolder->get($payslipsMatriculeDirPath);

                                                    // check if user payslip directory is not already shared
                                                    if (!in_array($payslipsMatriculeDirPathNode->getId(), $payslipMatriculeShareNodeIdList)) {
                                                        // share "MATRICULE" directory in read only with this agent user (found by email)
                                                        $this->_logCsvLine($this->_l10n->t('Partage du dossier SIRET-MATRICULE [' . $siretMatriculeDirName . '] en lecture seule.'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                        $payslipMatriculeShare = $this->_shareManager->newShare();
                                                        $payslipMatriculeShare->setNode($payslipsMatriculeDirPathNode);
                                                        $payslipMatriculeShare->setShareType(\OC\Share\Share::SHARE_TYPE_USER);
                                                        $payslipMatriculeShare->setSharedWith($userFindByEmailId);
                                                        $payslipMatriculeShare->setPermissions(\OCP\Constants::PERMISSION_READ);
                                                        $payslipMatriculeShare->setSharedBy($this->_userId);
                                                        $payslipMatriculeShare = $this->_shareManager->createShare($payslipMatriculeShare);
                                                    } else {
                                                        $this->_logCsvLine($this->_l10n->t('Dossier SIRET-MATRICULE [' . $siretMatriculeDirName . '] est déjà partagé.'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                    }

                                                    // "SIRET-MATRICULE/ANNEE" directory
                                                    $this->_logCsvLine($this->_l10n->t('Vérification de l\'existance du dossier de l\'annéee [' . $year . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                    try {
                                                        $userFolder->get($payslipsMatriculeYearDirPath);
                                                    } catch (Exception $e) {
                                                        $this->_logCsvLine($this->_l10n->t('Création du dossier de l\'annéee [' . $year . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                        $createdFolder = $userFolder->newFolder($payslipsMatriculeYearDirPath);
                                                        if (!$createdFolder) {
                                                            $archiveError++;
                                                            $errorMsg = $this->_l10n->t('Impossible de créer le dossier "' . $payslipsMatriculeYearDirPath . '[siret="' . $agentUser->getSiret() . '", matricule="' . $agentUser->getMatricule() . '", année="' . $year . '"].');
                                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                        }
                                                    }

                                                    if (!$this->hasError()) {
                                                        // move user agent payslip file in "SIRET-MATRICULE/ANNEE/PERIODE.pdf"
                                                        $pdfPayslipFilePath = $payslipsMatriculeYearDirPath . DIRECTORY_SEPARATOR . $period . '.pdf';
                                                        $this->_logCsvLine($this->_l10n->t('Déplacement du fichier PDF du bulletin de paie de la période [' . $period . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);
                                                        try {
                                                            $pdfPayslipCheckFileNode = $userFolder->get($pdfPayslipCheckFilePath);
                                                            $pdfPayslipCheckFileNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath);
                                                            $dematpayslipArchive->setIdFile($pdfPayslipCheckFileNode->getId());
                                                            $dematpayslipArchive->setFilePath($pdfPayslipCheckFileNode->getPath());
                                                            $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                                        } catch (Exception $e) {
                                                            $archiveError++;
                                                            $errorMsg = $this->_l10n->t('Impossible de déplacer le bulletin de paie "' . $pdfPayslipCheckFilePath . '" dans "' . $userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath . '".');
                                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                        }
                                                    }
                                                } else {
                                                    // agent disabled payslip
                                                    $this->_logCsvLine($this->_l10n->t('L\'utlisateur a désactivé la dématérialisation des fiches de paie [email=' . $agentUser->getEmail() . '].'), self::LOG_DEBUG, $agentUser, $startDate, $endDate);

                                                    // create PDF payslip
                                                    $payslipNoDematHorodateDirPath = $this->_payslipNoIndexDir . DIRECTORY_SEPARATOR . $phaseStartDateTime->format('Y-m-d_H-i-s');
                                                    try {
                                                        $userFolder->get($payslipNoDematHorodateDirPath);
                                                    } catch (Exception $e) {
                                                        $createdFolder = $userFolder->newFolder($payslipNoDematHorodateDirPath);
                                                        if (!$createdFolder) {
                                                            $archiveError++;
                                                            $errorMsg = $this->_l10n->t('Impossible de créeer le dossier "' . $payslipNoDematHorodateDirPath . '".');
                                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                        }
                                                    }

                                                    $payslipNoDematHorodateAgentDirPath = $payslipNoDematHorodateDirPath . DIRECTORY_SEPARATOR . $agentUser->getSiret() . '-' . $agentUser->getMatricule();
                                                    try {
                                                        $userFolder->get($payslipNoDematHorodateAgentDirPath);
                                                    } catch (Exception $e) {
                                                        $createdFolder = $userFolder->newFolder($payslipNoDematHorodateAgentDirPath);
                                                        if (!$createdFolder) {
                                                            $archiveError++;
                                                            $errorMsg = $this->_l10n->t('Impossible de créeer le dossier "' . $payslipNoDematHorodateAgentDirPath . '".');
                                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                        }
                                                    }
                                                    $payslipNoDematPDFPath = $payslipNoDematHorodateAgentDirPath . DIRECTORY_SEPARATOR . $startDate . '_' . $endDate;

                                                    // create PDF payslip in no index folder
                                                    $this->_logCsvLine($this->_l10n->t('Créer le PDF du bulletin de paie non dematérialisé.'), self::LOG_DEBUG);
                                                    $pdfPayslipFilePath = $payslipNoDematPDFPath . '.pdf';
                                                    try {
                                                        //$payslipNoDematPDFNode = $userFolder->get($payslipNoDematPDFPath);
                                                        $payslipNoDematPDFNode = $userFolder->get($pdfPayslipCheckFilePath);
                                                        $payslipNoDematPDFNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath);
                                                        $dematpayslipArchive->setIdFile($payslipNoDematPDFNode->getId());
                                                        $dematpayslipArchive->setFilePath($payslipNoDematPDFNode->getPath());
                                                        $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                                    } catch (Exception $e) {
                                                        $archiveError++;
                                                        $errorMsg = $this->_l10n->t('Impossible de déplacer le PDF du bulletin de paie non dematérialisé "' . $userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipFilePath . '".');
                                                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                                        $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                                    }
                                                }
                                            } else {
                                                $archiveError++;
                                                $errorMsg = $this->_l10n->t('L\'utilisateur "' . $userFindByEmailId . '" n\'a pas de dossier SIRET-MATRICULE associé [siret="' . $agentUser->getSiret() . '", matricule="' . $agentUser->getMatricule() . '", email="' . $agentUser->getEmail() . '"].');
                                                $this->_logCsvLine($errorMsg, self::LOG_ERROR, $agentUser, $startDate, $endDate);
                                                $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                            }
                                        }

                                        if ($archiveError > 0) {
                                            // file finished with errors
                                            $processMsgList['error'][] = $dematpayslipArchive;
                                            $dematpayslipArchive->setError(1);
                                            $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                        } else {
                                            $processMsgList['success'][] = $dematpayslipArchive;
                                            $dematpayslipArchive->setStatus(DematpayslipArchive::STATUS_INDEXED);
                                            $dematpayslipArchive->setStatusTime($phaseStartTime);
                                            // next operation for this file if asked
                                            if ($operationAskArchive) {
                                                $dematpayslipArchive->setOperationProgress(DematpayslipArchive::OPERATION_ARCHIVE);
                                            } else {
                                                $dematpayslipArchive->setOperationProgress(DematpayslipArchive::OPERATION_FINISHED);
                                            }
                                            $dematpayslipArchive->setError(0);
                                            $dematpayslipArchiveMapper->update($dematpayslipArchive);
                                        }
                                        if ($this->hasError()) {
                                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin du bulletin : ' . ($csvCheckLineNum + 1)) . ' --', self::LOG_ERROR);
                                            break;
                                        } else {
                                            $this->_logCsvLine('-- ' . $this->_l10n->t('Fin du bulletin : ' . ($csvCheckLineNum + 1)) . ' --', self::LOG_SUCCESS);
                                        }
                                    }
                                }
                            }

                            // create PDF for payslip not to be indexed (skip)
                            if ($pdfPayslipSkip !== null) {
                                $this->_logCsvLine($this->_l10n->t('Créer le PDF des bulletins de paie à ne pas indexer.'), self::LOG_DEBUG);
                                $pdfPayslipSkip->Output('F', $userHomeFilesDirPath . DIRECTORY_SEPARATOR . $payslipSkipPDFPath);
                                $pdfPayslipSkipPath = $payslipSkipPDFPath . '.pdf';
                                try {
                                    $payslipSkipPDFPathNode = $userFolder->get($payslipSkipPDFPath);
                                    $payslipSkipPDFPathNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipSkipPath);

                                    // update skip archive list
                                    if (!empty($skipArchiveList)) {
                                        foreach ($skipArchiveList as $skipArchive) {
                                            $processMsgList['success'][] = $skipArchive;
                                            // remove payslip created in check directory
                                            //try {
                                            //    $skipArchiveNode = $this->_rootFolder->get($skipArchive->getFilePath());
                                            //    $skipArchiveNode->delete();
                                            //} catch (Exception $e) {
                                            //    $errorMsg = $this->_l10n->t('Impossible de supprimer "' . $skipArchive->getFilePath() . '".');
                                            //    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                            //    $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                            //}

                                            $skipArchive->setIdFile($payslipSkipPDFPathNode->getId());
                                            $skipArchive->setFilePath($payslipSkipPDFPathNode->getPath());
                                            $skipArchive->setStatus(DematpayslipArchive::STATUS_INDEXED);
                                            $skipArchive->setStatusTime($phaseStartTime);
                                            $skipArchive->setOperationProgress(DematpayslipArchive::OPERATION_FINISHED); // no next operation
                                            $skipArchive->setError(0);
                                            $dematpayslipArchiveMapper->update($skipArchive);
                                        }
                                    }
                                } catch (Exception $e) {
                                    $errorMsg = $this->_l10n->t('Impossible de déplacer le PDF des bulletins de paie à ne pas indexer "' . $userFolder->getPath() . DIRECTORY_SEPARATOR . $pdfPayslipSkipPath . '".');
                                    $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                    $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $csvCheckLineNum);
                                }
                            }
                        }
                    }

                    if (!$this->hasError()) {
                        $phaseStep = 'report';
                        $phaseStepNum = 0;
                        $retryPhaseStepNum = $phaseStepNum;
                        if ($this->getPhaseStep() == $phaseStep) {
                            $retryPhaseStepNum = $this->getPhaseStepNum();
                        }
                        if (empty($this->getPhaseStep())) {
                            $reportSubDirName = $phaseStartDateTime->format('Y-m-d_H-i-s');
                            $reportSubDirPath = $this->_reportDirUserPath . DIRECTORY_SEPARATOR . $reportSubDirName;

                            // create report sub-directory
                            if ($phaseStepNum >= $retryPhaseStepNum) {
                                $this->_logCsvLine($this->_l10n->t('Création du dossier du rapport.'), self::LOG_DEBUG);
                                try {
                                    $userFolder->get($reportSubDirPath);
                                } catch (Exception $e) {
                                    $createdFolder = $userFolder->newFolder($reportSubDirPath);
                                    if (!$createdFolder) {
                                        $errorMsg = $this->_l10n->t('Impossible de créer le dossier du rapport "' . $reportSubDirName . '".');
                                        $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                        $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $phaseStepNum);
                                    }
                                }
                            }

                            if (!$this->hasError()) {
                                // move uploaded CSV file in report directory
                                $phaseStepNum++;
                                if ($phaseStepNum >= $retryPhaseStepNum) {
                                    $csvUploadFileName = $this->getCsvOrigin();
                                    if (!empty($csvUploadFileName)) {
                                        $this->_logCsvLine($this->_l10n->t('Déplacement du fichier CSV d\'upload dans le dossier du rapport.'), self::LOG_DEBUG);
                                        $csvUploadFileUserPath = $this->_uploadDirUserPath . DIRECTORY_SEPARATOR . $csvUploadFileName;
                                        $reportFileUserPath = $reportSubDirPath . DIRECTORY_SEPARATOR . $csvUploadFileName;
                                        try {
                                            $csvUploadFileUserNode = $userFolder->get($csvUploadFileUserPath);
                                            $csvUploadFileUserNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $reportFileUserPath);
                                        } catch (Exception $e) {
                                            $errorMsg = $this->_l10n->t('Impossible de déplacer le fichier CSV d\'upload "' . $csvUploadFileName . '" dans le dossier du rapport "' . $reportSubDirName . '".');
                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $phaseStepNum);
                                        }
                                    }
                                }

                                // move uploaded PDF file in report directory
                                $phaseStepNum++;
                                if ($phaseStepNum >= $retryPhaseStepNum) {
                                    $pdfUploadFileName = $this->getPdfOrigin();
                                    if (!empty($pdfUploadFileName)) {
                                        $this->_logCsvLine($this->_l10n->t('Déplacement du fichier PDF d\'upload dans le dossier du rapport.'), self::LOG_DEBUG);
                                        $pdfUploadFileUserPath = $this->_uploadDirUserPath . DIRECTORY_SEPARATOR . $pdfUploadFileName;
                                        $reportFileUserPath = $reportSubDirPath . DIRECTORY_SEPARATOR . $pdfUploadFileName;
                                        try {
                                            $pdfUploadFileUserNode = $userFolder->get($pdfUploadFileUserPath);
                                            $pdfUploadFileUserNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $reportFileUserPath);
                                        } catch (Exception $e) {
                                            $errorMsg = $this->_l10n->t('Impossible de déplacer le fichier PDF d\'upload "' . $pdfUploadFileName . '" dans le dossier du rapport "' . $reportSubDirName . '".');
                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $phaseStepNum);
                                        }
                                    }
                                }

                                // move CSV created in check step in report directory
                                $phaseStepNum++;
                                if ($phaseStepNum >= $retryPhaseStepNum) {
                                    if (!empty($csvCheckFileName)) {
                                        $this->_logCsvLine($this->_l10n->t('Déplacement du fichier CSV créé après vérification dans le dossier du rapport.'), self::LOG_DEBUG);
                                        $csvCheckFileUserPath = $this->_checkDirUserPath . DIRECTORY_SEPARATOR . $csvCheckFileName;
                                        $reportFileUserPath = $reportSubDirPath . DIRECTORY_SEPARATOR . $csvCheckFileName;
                                        try {
                                            $csvCheckFileUserNode = $userFolder->get($csvCheckFileUserPath);
                                            $csvCheckFileUserNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $reportFileUserPath);
                                        } catch (Exception $e) {
                                            $errorMsg = $this->_l10n->t('Impossible de déplacer le fichier CSV de vérification "' . $csvCheckFileName . '" dans le dossier du rapport "' . $reportSubDirName . '".');
                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $phaseStepNum);
                                        }
                                    }
                                }

                                // move created log file (during check process) in report directory
                                $phaseStepNum++;
                                if ($phaseStepNum >= $retryPhaseStepNum) {
                                    // get csv log file (during check process)
                                    $logCheckFileName = self::PHASE_ID_CHECK . '_log.csv';
                                    if (!empty($logCheckFileName)) {
                                        $this->_logCsvLine($this->_l10n->t('Déplacement du fichier de log créé après vérification dans le dossier du rapport.'), self::LOG_DEBUG);
                                        $logCheckFileUserPath = $this->_checkDirUserPath . DIRECTORY_SEPARATOR . $logCheckFileName;
                                        $reportFileUserPath = $reportSubDirPath . DIRECTORY_SEPARATOR . $logCheckFileName;
                                        try {
                                            $logCheckFileUserNode = $userFolder->get($logCheckFileUserPath);
                                            $logCheckFileUserNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $reportFileUserPath);
                                        } catch (Exception $e) {
                                            $errorMsg = $this->_l10n->t('Impossible de déplacer le fichier de log de vérification "' . $logCheckFileName . '" dans le dossier du rapport "' . $reportSubDirName . '".');
                                            $this->_logCsvLine($errorMsg, self::LOG_ERROR);
                                            $this->_processErrorMsg(-1, $errorMsg, $phaseStep, $phaseStepNum);
                                        }
                                    }
                                }
                            }
                        }
                    }
                }
                $successMsg = 'Opération "' . self::_getPhaseName($this->getPhase()) . '" terminée avec succès.';
                if (!$this->hasError()) {
                    $this->_logCsvLine($this->_l10n->t($successMsg), self::LOG_SUCCESS);
                }
                $this->_logCsvLine('## ' . $this->_l10n->t('Fin du processus d\'indexation.') . ' ##', self::LOG_DEBUG);
                fclose($this->_logFile);

                // rename log file
                try {
                    $logFileNode = $userFolder->get($logFilePath);
                    $logFileNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $logFilePath . '.csv');
                } catch (Exception $e) {
                    $this->_msgList['error'][] = $this->_l10n->t('Impossible de déplacer le fichier de log "' . $userFolder->getPath() . DIRECTORY_SEPARATOR . $logFilePath . '.csv' . '" .');
                }

                $this->setProcessing(0);
                $this->_dematpayslipProcessMapper->update($this);
                if (!$this->hasError()) {
                    $this->_msgList['success'][] = $successMsg;
                    // create new process for next time
                    $createdProcess = new DematpayslipProcess();
                    $createdProcess->_processInit();
                    $this->_dematpayslipProcessMapper->insert($createdProcess);
                }
            }

            if ($this->hasError()) {
                return -1;
            } else {
                return 1;
            }
        } catch (Exception $e) {
            $this->_msgList['error'][] = 'Exception ' . __METHOD__ . ' : error=' . $e->getMessage();
            return -1;
        }
    }

    /**
     * Agent list
     *
     * @throws  Exception
     */
    public function agentList() {
        $this->_initDematFolders();
    }

    /**
     * Agent CSV
     *
     * @throws  Exception
     */
    public function agentCsv() {
        $this->_initDematFolders();

        if (!$this->hasError()) {
            $dematpayslipUserList = $this->_dematpayslipUserMapper->findAll(array(array('disabled', 'desc'), array('siret'), array('matricule')));

            if (!empty($dematpayslipUserList)) {
                $dateTimeNow = new DateTime('now', $this->_timeZone);
                $userFolder = $this->_rootFolder->getUserFolder($this->_userId);
                $csvFileName = $dateTimeNow->format('Y-m-d_H-i-s') . '_agent';
                $csvFilePath = $this->_exportDirUserPath . DIRECTORY_SEPARATOR . $csvFileName;
                $csvFileFullPath = $this->_exportDirFullPath . DIRECTORY_SEPARATOR . $csvFileName;
                $csvFile = null;
                try {
                    $userFolder->newFile($csvFilePath);
                    $csvFile = fopen($csvFileFullPath, 'w+');
                } catch (Exception $e) {
                    $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier CSV "' . $csvFilePath . '".');
                    $this->_msgList['error'][] = $errorMsg;
                }

                if (!$this->hasError()) {
                    $csvDelimiter = $this->_csvDelimiter;
                    $csvEnclosure = $this->_csvEnclosure;

                    $csvFields = array();
                    $csvFields[] = $this->_l10n->t('Utilisateur');
                    $csvFields[] = $this->_l10n->t('Siret');
                    $csvFields[] = $this->_l10n->t('Matricule');
                    $csvFields[] = $this->_l10n->t('Email');
                    $csvFields[] = $this->_l10n->t('Papier ?');

                    $res = fputcsv($csvFile, $csvFields, $csvDelimiter, $csvEnclosure);
                    if (!$res) {
                        $errorMsg = $this->_l10n->t('Ecriture des colonnes dans le fichier CSV.');
                        $this->_msgList['error'][] = $errorMsg;
                    }

                    if (!$this->hasError()) {
                        foreach ($dematpayslipUserList as $dematpayslipUser) {
                            $agent = $this->_userManager->get($dematpayslipUser->getUserId());

                            $csvFields = array();
                            //$csvFields[] = $dematpayslipUser->getUserId();
                            $csvFields[] = $agent->getDisplayName();
                            $csvFields[] = $dematpayslipUser->getSiret();
                            $csvFields[] = $dematpayslipUser->getMatricule();
                            $csvFields[] = $dematpayslipUser->getEmail();
                            $csvFields[] = $dematpayslipUser->getDisabled();

                            $res = fputcsv($csvFile, $csvFields, $csvDelimiter, $csvEnclosure);
                            if (!$res) {
                                $errorMsg = $this->_l10n->t('Ecriture de la ligne dans le fichier CSV.');
                                $this->_msgList['error'][] = $errorMsg;
                                break;
                            }
                        }
                    }

                    $res = fclose($csvFile);
                    if (!$res) {
                        $errorMsg = $this->_l10n->t('Fermeture du fichier CSV.');
                        $this->_msgList['error'][] = $errorMsg;
                    }

                    if (!$this->hasError()) {
                        // rename csv file
                        try {
                            $csvFileNode = $userFolder->get($csvFilePath);
                            $csvFileNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $csvFilePath . '.csv');
                            $this->_msgList['success'][] = $this->_l10n->t('La liste des agents a bien été exporté au format CSV.');
                        } catch (Exception $e) {
                            $this->_msgList['error'][] = $this->_l10n->t('Impossible de déplacer le fichier CSV "' . $csvFilePath . '" .');
                        }
                    }
                }
            }
        }
    }

    /**
     * Archive list
     *
     * @throws Exception
     */
    public function archiveList() {
        $this->_initDematFolders();
    }

    /**
     * Archive action
     *
     * @param $action
     * @param $archiveSelectIdList
     * @return bool
     *
     * @throws Exception
     */
    public function archiveAction($action, $archiveSelectIdList) {
        $refresh = false;

        $this->_initDematFolders();

        if ($action == 'refresh_status') {
            // refresh all archive status
            $refresh = true;
        } elseif ($action == 'create') {
            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

            if (!empty($archiveSelectIdList)) {
                $processDoneIdList = array();
                foreach ($archiveSelectIdList as $archiveSelectId) {
                    $dematpayslipArchive = $dematpayslipArchiveMapper->find(intval($archiveSelectId));
                    if (!empty($dematpayslipArchive)) {
                        // get operation progress and launch process
                        $operationProgress = $dematpayslipArchive->getOperationProgress();
                        if (!in_array($dematpayslipArchive->getDematIdProcess(), $processDoneIdList) && $operationProgress == DematpayslipArchive::OPERATION_INDEXATION) {
                            $processList = $this->_dematpayslipProcessMapper->findAllLast();
                            if (!empty($processList)) {
                                $process = $processList[0];
                                if ($dematpayslipArchive->getDematIdProcess() == $process->getId() && $process->getPhase() == DematpayslipProcess::PHASE_ID_INDEXATION) {
                                    $processMsgList = array(
                                        'error'   => array(),
                                        'success' => array(),
                                    );
                                    $process->load($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $this->_shareManager, $this->_userManager, $this->_userId, $this->_dematpayslipProcessMapper, $this->_dematpayslipProcessSkipMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                                    $process->indexation($processMsgList);

                                    $processDoneIdList[] = $process->getId();
                                    $errorMsgList = $process->getMsgList('error');
                                    $successMsgList = $process->getMsgList('success');
                                    if (!empty($errorMsgList)) {
                                        $this->_msgList['error'][] = end($errorMsgList);
                                    }
                                    if (!empty($successMsgList)) {
                                        $this->_msgList['success'][] = end($successMsgList);
                                    }
                                }
                            }
                        } else if ($operationProgress == DematpayslipArchive::OPERATION_ARCHIVE) {
                            $dematpayslipArchive->archiveProcess($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $dematpayslipArchiveMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                            $errorMsgList = $dematpayslipArchive->getMsgList('error');
                            $successMsgList = $dematpayslipArchive->getMsgList('success');
                            if (!empty($errorMsgList)) {
                                $this->_msgList['error'][] = end($errorMsgList);
                            }
                            if (!empty($successMsgList)) {
                                $this->_msgList['success'][] = end($successMsgList);
                            }
                        }
                    }
                }
            }
        } elseif ($action == 'create_all') {
            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);
            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAllByOperationProgress(DematpayslipArchive::OPERATION_ARCHIVE);

            if (!empty($dematpayslipArchiveList)) {
                foreach ($dematpayslipArchiveList as $dematpayslipArchive) {
                    // only for operation ask
                    $operationAsk = $dematpayslipArchive->getOperationAsk();
                    $operationAskList = explode('+', $operationAsk);
                    if (in_array(DematpayslipArchive::OPERATION_ARCHIVE, $operationAskList)) {
                        $dematpayslipArchive->archiveProcess($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $dematpayslipArchiveMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                        $errorMsgList = $dematpayslipArchive->getMsgList('error');
                        $successMsgList = $dematpayslipArchive->getMsgList('success');
                        if (!empty($errorMsgList)) {
                            $this->_msgList['error'][] = end($errorMsgList);
                        }
                        if (!empty($successMsgList)) {
                            $this->_msgList['success'][] = end($successMsgList);
                        }
                    }
                }
            }
        } elseif ($action == 'run_again') {
            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

            if (!empty($archiveSelectIdList)) {
                $processDoneIdList = array();
                foreach ($archiveSelectIdList as $archiveSelectId) {
                    $dematpayslipArchive = $dematpayslipArchiveMapper->find(intval($archiveSelectId));
                    if (!empty($dematpayslipArchive)) {
                        // get operation progress and launch process
                        $operationProgress = $dematpayslipArchive->getOperationProgress();
                        if (!in_array($dematpayslipArchive->getDematIdProcess(), $processDoneIdList) && $operationProgress == DematpayslipArchive::OPERATION_INDEXATION) {
                            $processList = $this->_dematpayslipProcessMapper->findAllLast();
                            if (!empty($processList)) {
                                $process = $processList[0];
                                if ($dematpayslipArchive->getDematIdProcess() == $process->getId() && $process->getPhase() == DematpayslipProcess::PHASE_ID_INDEXATION) {
                                    $processMsgList = array(
                                        'error'   => array(),
                                        'success' => array(),
                                    );
                                    $process->load($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $this->_shareManager, $this->_userManager, $this->_userId, $this->_dematpayslipProcessMapper, $this->_dematpayslipProcessSkipMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                                    $process->indexation($processMsgList);

                                    $processDoneIdList[] = $process->getId();
                                    $errorMsgList = $process->getMsgList('error');
                                    $successMsgList = $process->getMsgList('success');
                                    if (!empty($errorMsgList)) {
                                        $this->_msgList['error'][] = end($errorMsgList);
                                    }
                                    if (!empty($successMsgList)) {
                                        $this->_msgList['success'][] = end($successMsgList);
                                    }
                                }
                            }
                        } else if ($operationProgress == DematpayslipArchive::OPERATION_ARCHIVE) {
                            $dematpayslipArchive->archiveProcess($this->_appName, $this->_config, $this->_dbConnection, $this->_l10n, $this->_rootFolder, $dematpayslipArchiveMapper, $this->_dematpayslipUserMapper, $this->_timeZone);
                            $errorMsgList = $dematpayslipArchive->getMsgList('error');
                            $successMsgList = $dematpayslipArchive->getMsgList('success');
                            if (!empty($errorMsgList)) {
                                $this->_msgList['error'][] = end($errorMsgList);
                            }
                            if (!empty($successMsgList)) {
                                $this->_msgList['success'][] = end($successMsgList);
                            }
                        }
                    }
                }
            }
        } elseif ($action == 'delete') {
            $nowDateTime = new DateTime('now', $this->_timeZone);
            $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);

            if (!empty($archiveSelectIdList)) {
                foreach ($archiveSelectIdList as $archiveSelectId) {
                    $dematpayslipArchive = $dematpayslipArchiveMapper->find(intval($archiveSelectId));
                    if (!empty($dematpayslipArchive)) {
                        if (!empty($dematpayslipArchive->getPastellIdDocument())) {
                            $this->_msgList['error'][] = $this->_l10n->t('Impossible de supprimer le bulletin de paie "' . $dematpayslipArchive->getFilePath() . '" déjà en cours d\'archivage.');;
                        } else {
                            $dematpayslipArchive->setStatus(DematpayslipArchive::STATUS_DELETED);
                            $dematpayslipArchive->setStatusTime($nowDateTime->getTimestamp());
                            $dematpayslipArchiveMapper->update($dematpayslipArchive);
                        }
                    }

                    if ($this->hasError()) break;
                }
            }
        }

        return $refresh;
    }
    
    /**
     * Archive CSV
     *
     * @param   int         $stepStatus     Step status
     * 
     * @throws  Exception
     */
    public function archiveCsv($stepStatus) {
        $this->_initDematFolders();

        $dematpayslipArchiveMapper = new DematpayslipArchiveMapper($this->_dbConnection);
        if ($stepStatus === -1) {
            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAll(0);
        } else {
            $dematpayslipArchiveList = $dematpayslipArchiveMapper->findAllByStepStatus($stepStatus);
        }

        if (!empty($dematpayslipArchiveList)) {
            $dateTimeNow = new DateTime('now', $this->_timeZone);
            $userFolder = $this->_rootFolder->getUserFolder($this->_userId);
            $csvFileName = $dateTimeNow->format('Y-m-d_H-i-s') . '_archive';
            $csvFilePath = $this->_exportDirUserPath . DIRECTORY_SEPARATOR . $csvFileName;
            $csvFileFullPath = $this->_exportDirFullPath . DIRECTORY_SEPARATOR . $csvFileName;
            $csvFile = null;
            try {
                $userFolder->newFile($csvFilePath);
                $csvFile = fopen($csvFileFullPath, 'w+');
            } catch (Exception $e) {
                $errorMsg = $this->_l10n->t('Impossible d\'ouvrir le fichier CSV "' . $csvFilePath . '".');
                $this->_msgList['error'][] = $errorMsg;
            }

            if (!$this->hasError()) {
                $csvDelimiter = $this->_csvDelimiter;
                $csvEnclosure = $this->_csvEnclosure;

                $csvFields = array();
                $csvFields[] = $this->_l10n->t('Chemin');
                $csvFields[] = $this->_l10n->t('Siret');
                $csvFields[] = $this->_l10n->t('Matricule');
                $csvFields[] = $this->_l10n->t('Nom');
                $csvFields[] = $this->_l10n->t('Prénom');
                $csvFields[] = $this->_l10n->t('Début');
                $csvFields[] = $this->_l10n->t('Fin');
                $csvFields[] = $this->_l10n->t('Edition');
                $csvFields[] = $this->_l10n->t('Archive');
                $csvFields[] = $this->_l10n->t('Dernière action');
                $csvFields[] = $this->_l10n->t('Opérations demandées');
                $csvFields[] = $this->_l10n->t('Opération en cours');
                $csvFields[] = $this->_l10n->t('Etat');
                $csvFields[] = $this->_l10n->t('Erreur');

                $res = fputcsv($csvFile, $csvFields, $csvDelimiter, $csvEnclosure);
                if (!$res) {
                    $errorMsg = $this->_l10n->t('Ecriture des colonnes dans le fichier CSV.');
                    $this->_msgList['error'][] = $errorMsg;
                }

                if (!$this->hasError()) {
                    foreach ($dematpayslipArchiveList as $dematpayslipArchive) {
                        $csvFields = array();
                        $csvFields[] = $dematpayslipArchive->getFilePath();
                        $csvFields[] = $dematpayslipArchive->getSiret();
                        $csvFields[] = $dematpayslipArchive->getMatricule();
                        $csvFields[] = $dematpayslipArchive->getLastname();
                        $csvFields[] = $dematpayslipArchive->getFirstname();
                        $csvFields[] = $dematpayslipArchive->getStartDate();
                        $csvFields[] = $dematpayslipArchive->getEndDate();
                        $csvFields[] = $dematpayslipArchive->getEditionDate();
                        $csvFields[] = $dematpayslipArchive->getPastellIdDocument();
                        $csvFields[] = $dematpayslipArchive->getPastellLastAction();
                        $csvFields[] = $dematpayslipArchive->showOperationAskLabel();
                        $csvFields[] = $dematpayslipArchive->showOperationProgressLabel();
                        $csvFields[] = $dematpayslipArchive->showStatusLabel();
                        $csvFields[] = $dematpayslipArchive->getError();

                        $res = fputcsv($csvFile, $csvFields, $csvDelimiter, $csvEnclosure);
                        if (!$res) {
                            $errorMsg = $this->_l10n->t('Ecriture de la ligne dans le fichier CSV.');
                            $this->_msgList['error'][] = $errorMsg;
                            break;
                        }
                    }
                }

                $res = fclose($csvFile);
                if (!$res) {
                    $errorMsg = $this->_l10n->t('Fermeture du fichier CSV.');
                    $this->_msgList['error'][] = $errorMsg;
                }

                if (!$this->hasError()) {
                    // rename csv file
                    try {
                        $csvFileNode = $userFolder->get($csvFilePath);
                        $csvFileNode->move($userFolder->getPath() . DIRECTORY_SEPARATOR . $csvFilePath . '.csv');
                        $this->_msgList['success'][] = $this->_l10n->t('La liste des archives a bien été exporté au format CSV.');
                    } catch (Exception $e) {
                        $this->_msgList['error'][] = $this->_l10n->t('Impossible de déplacer le fichier CSV "' . $csvFilePath . '" .');
                    }
                }
            }
        }
    }

    /**
     * User list
     *
     * @param $userEmail
     * @return array
     */
    public function userList($userEmail) {
        $userList = array();

        if (!empty($userEmail)) {
            // /!\ remove all users not available (FIX for multiple users with SAML)
            if (DematpayslipProcess::USER_REMOVE_UNKNOWN == 1) {
                $userList = DematpayslipProcess::_removeUnknownUser($this->_userManager->getByEmail($userEmail));
            }
        }

        return $userList;
    }

    /**
     * Shares (payslip user list)
     *
     * @param   int             $userId     User id
     * @return  array|Entity[]
     */
    public function shares($userId) {
        $dematpayslipUserList = $this->_dematpayslipUserMapper->findAllByUserId($userId);
        $nbDematpayslipUserList = count($dematpayslipUserList);
        if ($nbDematpayslipUserList <= 0) {
            $this->_msgList['error'][] = $this->_l10n->t('Cet utilisateur n\'a pas encore de bulletin de paie dématérialisé.');
        }

        return $dematpayslipUserList;
    }
}
